# -*- coding: UTF-8 -*-
"""
Provide standard ways to get various dirs

This is similar to th pyxdg module but it does not automatically creates the
dirs. Not creating the dirs is important to be able to show default values in
`yokadid --help` output without creating anything.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os
import getpass
import shutil
import tempfile

from yokadi.core import fileutils


_WINDOWS = os.name == "nt"

DB_NAME = "yokadi.db"


class MigrationException(Exception):
    pass


def _getAppDataDir():
    assert _WINDOWS
    return os.environ["APPDATA"]


def getRuntimeDir():
    value = os.getenv("XDG_RUNTIME_DIR")
    if not value:
        # Running on a system where XDG_RUNTIME_DIR is not set, fallback to
        # $tempdir/yokadi-$user
        tmpdir = tempfile.gettempdir()
        value = os.path.join(tmpdir, "yokadi-" + getpass.getuser())
    return value


def getLogDir():
    return getCacheDir()


def getCacheDir():
    if _WINDOWS:
        value = os.path.join(_getAppDataDir(), "yokadi", "cache")
    else:
        cacheBaseDir = os.getenv("XDG_CACHE_HOME")
        if not cacheBaseDir:
            cacheBaseDir = os.path.expandvars("$HOME/.cache")
        value = os.path.join(cacheBaseDir, "yokadi")
    return value


def getDataDir():
    xdgDataDir = os.environ.get("XDG_DATA_HOME")
    if xdgDataDir:
        return os.path.join(xdgDataDir, "yokadi")

    if _WINDOWS:
        return os.path.join(_getAppDataDir(), "yokadi", "data")

    return os.path.expandvars("$HOME/.local/share/yokadi")


def getHistoryPath():
    path = os.getenv("YOKADI_HISTORY")
    if path:
        return path
    return os.path.join(getCacheDir(), "history")


def getDbPath(dataDir):
    path = os.getenv("YOKADI_DB")
    if path:
        return path
    return os.path.join(dataDir, "yokadi.db")


def _getOldHistoryPath():
    if _WINDOWS:
        return os.path.join(_getAppDataDir(), ".yokadi_history")
    else:
        return os.path.expandvars("$HOME/.yokadi_history")


def migrateOldHistory():
    oldHistoryPath = _getOldHistoryPath()
    if not os.path.exists(oldHistoryPath):
        return

    newHistoryPath = getHistoryPath()
    if os.path.exists(newHistoryPath):
        # History is not critical, just overwrite the new file
        os.unlink(newHistoryPath)
    fileutils.createParentDirs(newHistoryPath)
    shutil.move(oldHistoryPath, newHistoryPath)
    print("Moved %s to %s" % (oldHistoryPath, newHistoryPath))


def migrateOldDb(newDbPath):
    oldDbPath = os.path.normcase(os.path.expandvars("$HOME/.yokadi.db"))
    if not os.path.exists(oldDbPath):
        return

    if os.path.exists(newDbPath):
        raise MigrationException("Tried to move %s to %s, but %s already exists."
                                 " You must remove one of the two files." % (oldDbPath, newDbPath, newDbPath))
    fileutils.createParentDirs(newDbPath)
    shutil.move(oldDbPath, newDbPath)
    print("Moved %s to %s" % (oldDbPath, newDbPath))



# -*- coding: UTF-8 -*-
"""
Help test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os
import sys
import unittest

from cmd import Cmd
from contextlib import contextmanager

from yokadi.core import db
from yokadi.ycli.main import YokadiCmd


@contextmanager
def to_devnull(out):
    out_fd = out.fileno()
    with open(os.devnull, "wb") as null, \
            os.fdopen(os.dup(out_fd), "wb") as copied:
        out.flush()
        os.dup2(null.fileno(), out_fd)
        try:
            yield
        finally:
            out.flush()
            os.dup2(copied.fileno(), out_fd)


class HelpTestCase(unittest.TestCase):
    """
    A basic test for the command helps: it just execute 'help <cmd>' on all
    commands. This catches invalid format characters in the help strings.
    """
    def setUp(self):
        # Some help commands look into the db for default values
        db.connectDatabase("", memoryDatabase=True)
        db.setDefaultConfig()

    def testHelp(self):
        cmd = YokadiCmd()
        for attr in dir(cmd):
            if not attr.startswith("do_"):
                continue

            yokadiCommand = attr[3:]
            try:
                # Execute the command, but redirect stdout and stderr to
                # /dev/null to avoid flooding the terminal
                with to_devnull(sys.stdout), to_devnull(sys.stderr):
                    # We use Cmd implementation of onecmd() because YokadiCmd
                    # overrides it to catch exceptions
                    Cmd.onecmd(cmd, "help " + yokadiCommand)
            except Exception:
                print("'help %s' failed" % yokadiCommand)
                raise

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
HTML rendering of t_list output

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import xml.sax.saxutils as saxutils

from collections import namedtuple

TaskField = namedtuple("TaskField", ("title", "format"))


HTML_HEADER = """
<html lang='en'>
<head>
    <meta charset='utf-8'>
    <meta name='viewport' content='width=device-width, initial-scale=1, shrink-to-fit=no'>
    <style>
        table {
            border-collapse: collapse;
        }
        td, th {
            border: 1px solid #ccc;
            padding: 0.5em;
            vertical-align: top;
        }
    </style>
    <title>Yokadi tasks export</title>
</head>
<body>
"""


HTML_FOOTER = "</body></html>"


def escape(text):
    return saxutils.escape(str(text))


def printRow(out, tag, lst):
    print("<tr>", file=out)
    for value in lst:
        if value:
            text = escape(value).replace("\n", "<br>")
        else:
            text = "&nbsp;"
        print("<%s>%s</%s>" % (tag, text, tag), file=out)
    print("</tr>", file=out)


class HtmlListRenderer(object):
    def __init__(self, out):
        self.out = out

        print(HTML_HEADER, file=self.out)

    def addTaskList(self, sectionName, taskList):
        """Store tasks for this section
        @param sectionName: name of the task groupement section
        @type sectionName: unicode
        @param taskList: list of tasks to display
        @type taskList: list of db.Task instances
        """
        TASK_FIELDS = [
            TaskField("Id", lambda x: str(x.id)),
            TaskField("Title", self._titleFormater),
            TaskField("Due date", lambda x: str(x.dueDate)),
            TaskField("Urgency", lambda x: str(x.urgency)),
            TaskField("Status", lambda x: x.status),
        ]

        print("<h1>%s</h1>" % escape(sectionName), file=self.out)
        print("<table width='100%'>", file=self.out)
        printRow(self.out, "th", [x.title for x in TASK_FIELDS])
        for task in taskList:
            lst = [x.format(task) for x in TASK_FIELDS]
            printRow(self.out, "td", lst)
        print("</table>", file=self.out)

    def end(self):
        print(HTML_FOOTER, file=self.out)

    def _titleFormater(self, task):
        title = task.title
        keywords = task.getKeywordsAsString()
        if keywords:
            title += " " + keywords
        if task.description:
            title += "\n" + task.description
        return title
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Text rendering of t_list output

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""
from datetime import datetime, timedelta

import yokadi.ycli.colors as C
from yokadi.core import ydateutils
from yokadi.ycli import tui


VLINE = "│"
HLINE = "─"
CROSS = "┼"

LINE_COLOR = C.CYAN


def colorizer(value, reverse=False):
    """Return a color according to value.
    @param value: value used to determine color. Low (0) value means not urgent/visible, high (100) value means
                  important
    @param reverse: If false low value means important and vice versa
    @return: a color code or None for no color"""
    if reverse:
        value = 100 - value
    if value > 75:
        return C.RED
    elif value > 50:
        return C.PURPLE
    elif value > 25:
        return C.ORANGE
    else:
        return None


class Column(object):
    __slots__ = ["title", "width", "formater"]

    def __init__(self, title, width, formater):
        """
        formater is a callable which accepts a task and returns a tuple
        of the form (string, color)
        color may be None if no color should be applied
        """
        self.title = title
        self.width = width
        self.formater = formater

    def createHeader(self):
        return self.title.ljust(self.width)

    def createCell(self, task):
        value, color = self.formater(task)

        if color:
            cell = color + value.ljust(self.width) + C.RESET
        else:
            cell = value.ljust(self.width)

        return cell


def idFormater(task):
    return str(task.id), None


class TitleFormater(object):
    def __init__(self, width):
        self.width = width

    def __call__(self, task):
        colorizer = tui.TextColorizer()
        keywords = task.getUserKeywordsNameAsString()
        hasDescription = task.description is not None and task.description != ""

        maxWidth = self.width
        if hasDescription:
            maxWidth -= 1

        # Create title
        title = task.title
        if keywords and len(title) < maxWidth:
            title += ' '
            colorizer.setColorAt(len(title), C.ORANGE)
            title += keywords
            colorizer.setResetAt(len(title))

        # Crop title to fit in self.width
        titleWidth = len(title)
        if titleWidth > maxWidth:
            title = title[:maxWidth - 1] + ">"
            colorizer.crop(maxWidth - 1)
            colorizer.setResetAt(maxWidth - 1)
        else:
            title = title.ljust(maxWidth)

        if hasDescription:
            title = title + "*"

        title = colorizer.render(title)

        return title, None


def urgencyFormater(task):
    return str(task.urgency), colorizer(task.urgency)


def statusFormater(task):
    if task.status == "started":
        color = C.BOLD
    else:
        color = None
    return task.status[0].upper(), color


class AgeFormater(object):
    def __init__(self, today, asDate=False):
        self.today = today
        self.asDate = asDate

    def __call__(self, task):
        delta = self.today - task.creationDate.replace(microsecond=0)
        if self.asDate:
            return task.creationDate.strftime("%x %H:%M"), None
        else:
            return ydateutils.formatTimeDelta(delta), colorizer(delta.days)


class DueDateFormater(object):
    def __init__(self, today, shortFormat):
        self.today = today
        self.shortFormat = shortFormat

    def __call__(self, task):
        if not task.dueDate:
            return "", None
        delta = task.dueDate - self.today
        if delta.days != 0:
            value = task.dueDate.strftime("%x %H:%M")
        else:
            value = task.dueDate.strftime("%H:%M")

        if self.shortFormat:
            value = ydateutils.formatTimeDelta(delta)
        else:
            value += " (%s)" % ydateutils.formatTimeDelta(delta)

        color = colorizer(delta.days * 33, reverse=True)
        return value, color


class TextListRenderer(object):
    def __init__(self, out, termWidth=None, renderAsNotes=False, splitOnDate=False):
        """
        @param out: output target
        @param termWidth: terminal width (int)
        @param renderAsNotes: whether to display task as notes (with dates) instead of tasks (with age). (boot)"""
        self.out = out
        self.termWidth = termWidth or tui.getTermWidth()
        self.taskLists = []
        self.maxTitleWidth = len("Title")
        self.today = datetime.today().replace(microsecond=0)
        self.firstHeader = True
        self.splitOnDate = splitOnDate

        if self.termWidth < 100:
            dueColumnWidth = 8
            shortDateFormat = True
        else:
            dueColumnWidth = 26
            shortDateFormat = False

        if renderAsNotes:
            self.splitOnDate = True
            creationDateColumnWidth = 16
            creationDateTitle = "Creation date"
        else:
            creationDateColumnWidth = 8
            creationDateTitle = "Age"

        # All fields set to None must be defined in end()
        self.columns = [
            Column("ID", None, idFormater),
            Column("Title", None, None),
            Column("U", 3, urgencyFormater),
            Column("S", 1, statusFormater),
            Column(creationDateTitle, creationDateColumnWidth, AgeFormater(self.today, renderAsNotes)),
            Column("Due date", dueColumnWidth, DueDateFormater(self.today, shortDateFormat)),
        ]

        self.idColumn = self.columns[0]
        self.titleColumn = self.columns[1]

        self.maxId = 0

    def addTaskList(self, sectionName, taskList):
        """Store tasks for this section
        @param sectionName: name of the task groupment section
        @type sectionName: unicode
        @param taskList: list of tasks to display
        @type taskList: list of db.Task instances
        """
        self.taskLists.append((sectionName, taskList))
        # Find max title width
        for task in taskList:
            title = task.title
            keywords = task.getUserKeywordsNameAsString()
            if keywords:
                title = "{} ({})".format(title, keywords)
            titleWidth = len(title)
            if task.description:
                titleWidth += 1
            self.maxTitleWidth = max(self.maxTitleWidth, titleWidth)
            self.maxId = max(self.maxId, task.id)

    def end(self):
        today = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        # Adjust idColumn
        self.idColumn.width = max(2, len(str(self.maxId)))

        # Adjust titleColumn
        self.titleColumn.width = self.maxTitleWidth
        totalWidth = sum([x.width for x in self.columns]) + len(self.columns) - 1
        if totalWidth >= self.termWidth:
            self.titleColumn.width = self.termWidth - (totalWidth - self.titleColumn.width)
        self.titleColumn.formater = TitleFormater(self.titleColumn.width)

        # Print table
        for sectionName, taskList in self.taskLists:
            dateSplitters = [(1, "day"), (7, "week"), (30, "month"), (30 * 4, "quarter"), (365, "year")]
            splitterRange, splitterName = dateSplitters.pop()
            splitterText = None
            self._renderTaskListHeader(sectionName)
            for task in taskList:
                while self.splitOnDate and task.creationDate > today - timedelta(splitterRange):
                    splitterText = "Last %s" % splitterName
                    if len(dateSplitters) > 0:
                        splitterRange, splitterName = dateSplitters.pop()
                    else:
                        self.splitOnDate = False

                if splitterText:
                    print(C.GREEN + splitterText.center(totalWidth) + C.RESET, file=self.out)
                    splitterText = None

                self._renderTaskListRow(task)

    def _renderTaskListHeader(self, sectionName):
        """
        @param sectionName: name used for list header
        @type sectionName: unicode"""

        cells = [x.createHeader() for x in self.columns]
        width = sum([len(x) for x in cells]) + len(cells) - 1
        if self.firstHeader:
            self.firstHeader = False
        else:
            print(file=self.out)

        # section name
        print(C.CYAN + sectionName.center(width) + C.RESET, file=self.out)

        # header titles
        line = (LINE_COLOR + VLINE + C.RESET).join(cells)
        print(line, file=self.out)

        # header separator line
        cells = [HLINE * len(x) for x in cells]
        print(LINE_COLOR + CROSS.join(cells) + C.RESET, file=self.out)

    def _renderTaskListRow(self, task):
        cells = [column.createCell(task) for column in self.columns]
        sep = LINE_COLOR + VLINE + C.RESET
        print(sep.join(cells), file=self.out)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
TextListRenderer test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest
from io import StringIO

from yokadi.ycli import colors
from yokadi.core import dbutils

from yokadi.ycli import tui
from yokadi.ycli.textlistrenderer import TextListRenderer, TitleFormater
from yokadi.core import db


def stripColor(text):
    for colorcode in (colors.BOLD, colors.RED, colors.GREEN, colors.ORANGE, colors.PURPLE, colors.CYAN, colors.GREY,
                      colors.RESET):
        text = text.replace(colorcode, '')
    return text


class TextListRendererTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        tui.clearInputAnswers()

    def testTitleFormater(self):
        dbutils.getOrCreateProject("x", interactive=False)
        dbutils.getOrCreateKeyword("key1", interactive=False)
        dbutils.getOrCreateKeyword("key2", interactive=False)
        task = dbutils.addTask("x", "t1", {})
        taskWithKeywords = dbutils.addTask("x", "t2", {"key1": None, "key2": 12})

        longTask = dbutils.addTask("x", "01234567890123456789", {})
        longTask.description = "And it has a description"

        TEST_DATA = (
            (task, 20, "t1"),
            (taskWithKeywords, 20, "t2 @key1 @key2"),
            (taskWithKeywords, 4, "t2 >"),
            (longTask, 10, longTask.title[:8] + ">*"),
            (longTask, len(longTask.title), longTask.title[:-2] + ">*"),
            (longTask, len(longTask.title) + 1, longTask.title + "*"),
            (longTask, 40, longTask.title.ljust(39) + "*"),
        )

        for task, width, expected in TEST_DATA:
            with self.subTest(task=task, width=width):
                formater = TitleFormater(width)
                out = formater(task)[0]
                out = stripColor(out)

                expected = expected.ljust(width)
                self.assertEqual(out, expected)

    def testFullRendering(self):
        dbutils.getOrCreateProject("x", interactive=False)
        dbutils.getOrCreateKeyword("k1", interactive=False)
        dbutils.getOrCreateKeyword("k2", interactive=False)
        dbutils.addTask("x", "t1", {})
        t2 = dbutils.addTask("x", "t2", {"k1": None, "k2": 12})
        longTask = dbutils.addTask("x", "A longer task name", {})
        longTask.description = "And it has a description"

        out = StringIO()
        renderer = TextListRenderer(out, termWidth=80)
        renderer.addTaskList("Foo", [t2, longTask])
        self.assertEqual(renderer.maxTitleWidth, len(longTask.title) + 1)
        renderer.end()
        out = stripColor(out.getvalue())

        expected = \
            "                     Foo                      \n" \
            "ID│Title              │U  │S│Age     │Due date\n" \
            "──┼───────────────────┼───┼─┼────────┼────────\n" \
            "2 │t2 @k1 @k2         │0  │N│0m      │        \n" \
            "3 │A longer task name*│0  │N│0m      │        \n"
        self.assertMultiLineEqual(out, expected)


# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Date utilities.

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import operator
from datetime import date, datetime, timedelta

from yokadi.ycli import basicparseutils
from yokadi.core.yokadiexception import YokadiException

WEEKDAYS = {"monday": 0, "tuesday": 1, "wednesday": 2, "thursday": 3, "friday": 4, "saturday": 5, "sunday": 6}
SHORT_WEEKDAYS = {"mo": 0, "tu": 1, "we": 2, "th": 3, "fr": 4, "sa": 5, "su": 6}

TIME_HINT_BEGIN = "begin"
TIME_HINT_END = "end"

DATE_FORMATS = [
    "%d/%m/%Y",
    "%d/%m/%y",
    "%d/%m",
]

TIME_FORMATS = [
    "%H:%M:%S",
    "%H:%M",
    "%H",
]


def parseDateTimeDelta(line):
    # FIXME: Do we really want to support float deltas?
    try:
        delta = float(line[:-1])
    except ValueError:
        raise YokadiException("Timeshift must be a float or an integer")

    suffix = line[-1].upper()
    if suffix == "W":
        return timedelta(days=delta * 7)
    elif suffix == "D":
        return timedelta(days=delta)
    elif suffix == "H":
        return timedelta(hours=delta)
    elif suffix == "M":
        return timedelta(minutes=delta)
    else:
        raise YokadiException("Unable to understand time shift. See help t_set_due")


def testFormats(text, formats):
    for fmt in formats:
        try:
            return datetime.strptime(text, fmt), fmt
        except ValueError:
            pass
    return None, None


def guessTime(text):
    afternoon = False
    # We do not use the "%p" format to handle AM/PM because its behavior is
    # locale-dependent
    suffix = text[-2:]
    if suffix == "am":
        text = text[:-2].strip()
    elif suffix == "pm":
        afternoon = True
        text = text[:-2].strip()

    out, fmt = testFormats(text, TIME_FORMATS)
    if out is None:
        return None
    if afternoon:
        out += timedelta(hours=12)
    return out.time()


def parseHumaneDateTime(line, hint=None, today=None):
    """Parse human date and time and return structured datetime object
    Datetime  can be absolute (23/10/2008 10:38) or relative (+5M, +3H, +1D, +6W)
    @param line: human date / time
    @param hint: optional hint to tell whether time should be set to the
    beginning or the end of the day when not specified.
    @param today: optional parameter to define a fake today date. Useful for
    unit testing.
    @type line: str
    @return: datetime object"""
    def guessDate(text):
        out, fmt = testFormats(text, DATE_FORMATS)
        if not out:
            return None
        if "%y" not in fmt and "%Y" not in fmt:
            out = out.replace(year=today.year)
        return out.date()

    def applyTimeHint(date, hint):
        if not hint:
            return date
        if hint == TIME_HINT_BEGIN:
            return date.replace(hour=0, minute=0, second=0)
        elif hint == TIME_HINT_END:
            return date.replace(hour=23, minute=59, second=59)
        else:
            raise Exception("Unknown hint %s" % hint)

    line = basicparseutils.simplifySpaces(line).lower()
    if not line:
        raise YokadiException("Date is empty")

    if today is None:
        today = datetime.today().replace(microsecond=0)

    if line == "now":
        return today

    if line == "today":
        return applyTimeHint(today, hint)

    # Check for "+<delta>" format
    if line.startswith("+"):
        return today + parseDateTimeDelta(line[1:])
    if line.startswith("-"):
        return today - parseDateTimeDelta(line[1:])

    # Check for "<weekday> [<time>]" format
    firstWord = line.split()[0]

    weekdayDict = {
        "today": today.weekday(),
        "tomorrow": (today.weekday() + 1) % 7,
    }
    weekdayDict.update(WEEKDAYS)
    weekdayDict.update(SHORT_WEEKDAYS)
    weekday = weekdayDict.get(firstWord)
    if weekday is not None:
        date = today + timedelta(days=(weekday - today.weekday()) % 7)
        if " " in line:
            timeText = line.split(' ', 1)[1]
            tTime = guessTime(timeText)
            if tTime is None:
                raise YokadiException("Unable to understand time '%s'" % timeText)
            date = datetime.combine(date, tTime)
        else:
            date = applyTimeHint(date, hint)
        return date

    if " " in line:
        # Absolute date and time?
        dateText, timeText = line.split(' ', 1)
        tDate = guessDate(dateText)
        if tDate is not None:
            tTime = guessTime(timeText)
            if tTime is not None:
                return datetime.combine(tDate, tTime)

    # Only date?
    tDate = guessDate(line)
    if tDate is not None:
        dt = datetime.combine(tDate, today.time())
        return applyTimeHint(dt, hint)

    # Only time?
    tTime = guessTime(line)
    if tTime is not None:
        tDate = datetime.combine(today.date(), tTime)
        if tTime > today.time():
            return tDate
        else:
            return tDate + timedelta(days=1)

    raise YokadiException("Unable to understand date '%s'" % line)


def formatTimeDelta(delta):
    """Friendly format a time delta:
        - Show only days if delta > 1 day
        - Show only hours and minutes otherwise
    @param timeLeft: Remaining time
    @type timeLeft: timedelta (from datetime)
    @return: formated  str"""
    prefix = ""
    value = ""
    if delta < timedelta(0):
        delta = -delta
        prefix = "-"

    if delta.days >= 365:
        value = "%dY" % (delta.days / 365)
        days = delta.days % 365
        if days > 30:
            value += ", %dM" % (days / 30)
    elif delta.days > 50:
        value = "%dM" % (delta.days / 30)
        days = delta.days % 30
        if days > 0:
            value += ", %dd" % days
    elif delta.days >= 7:
        value = "%dw" % (delta.days / 7)
        days = delta.days % 7
        if days > 0:
            value += ", %dd" % days
    elif delta.days > 0:
        value = "%dd" % delta.days
    else:
        minutes = delta.seconds / 60
        hours = minutes / 60
        minutes = minutes % 60
        if hours >= 1:
            value = "%dh " % hours
        else:
            value = ""
        value += "%dm" % minutes

    return prefix + value


def getHourAndMinute(token):
    """Extract hour and minute from HH:MM token
    #TODO: move this in date utils
    @param token: HH:MM string
    @return: (int, int)"""
    try:
        hour, minute = token.split(":")
    except ValueError:
        hour = token
        minute = 0
    try:
        hour = int(hour)
        minute = int(minute)
    except ValueError:
        raise YokadiException("You must provide integer for hour/minute")
    return hour, minute


def getWeekDayNumberFromDay(day):
    """Return week day number (0-6) from week day name (short or long)
    @param day: week day as a string in short or long format (in english)
    @type day: str
    @return: week day number (int)"""
    if len(day) == 2 and day in SHORT_WEEKDAYS:
        dayNumber = SHORT_WEEKDAYS[day]
    elif day in WEEKDAYS:
        dayNumber = WEEKDAYS[day]
    else:
        raise YokadiException("Day must be one of the following: [mo]nday, [tu]esday, [we]nesday, [th]ursday, [fr]iday,"
                              " [sa]turday, [su]nday")
    return dayNumber


def parseDateLimit(line, today=None):
    """Parse a string of the form <operator><limit>
    - operator is one of: < <= >= > (default to <=)
    - limit is a date as understood by parseHumaneDateTime()

    @param line: the string to parse
    @param today: optional specification of current day, for unit testing
    @return: (operator, date)"""

    # Order matters: match longest operators first!
    operators = [
        ("<=", operator.__le__, TIME_HINT_END),
        (">=", operator.__ge__, TIME_HINT_BEGIN),
        (">", operator.__gt__, TIME_HINT_END),
        ("<", operator.__lt__, TIME_HINT_BEGIN),
    ]

    op = operator.__le__
    hint = TIME_HINT_END
    for txt, loopOp, loopHint in operators:
        if line.startswith(txt):
            op = loopOp
            hint = loopHint
            line = line[len(txt):]
            break

    limit = parseHumaneDateTime(line, today=today, hint=hint)
    return op, limit


def parseMinDate(line):
    # Parse the line string and return a minimum date
    minDate = date.today()
    if line == "today":
        pass
    elif line == "thisweek":
        minDate -= timedelta(minDate.weekday())
    elif line == "thismonth":
        minDate = minDate.replace(day=1)
    else:
        minDate = parseHumaneDateTime(line).date()

    return minDate

# vi: ts=4 sw=4 et



"""
Handling of common command line arguments

@author: Aurelien Gateau <mail@agateau.com>
@license: GPL v3 or later
"""
import os
import sys
import yokadi
from yokadi.core import basepaths
from yokadi.ycli import tui


def addArgs(parser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--datadir', dest='dataDir', help='Database dir (default: %s)' % basepaths.getDataDir(),
                       metavar='DATADIR')
    group.add_argument('-d', '--db', dest='dbPath',
                       help='TODO database (default: {}). This option is deprecated and will be removed in the next'
                       ' version of Yokadi. Use --datadir instead.'
                       .format(os.path.join('$DATADIR', basepaths.DB_NAME)),
                       metavar='FILE')
    parser.add_argument('-v', '--version', dest='version', action='store_true', help='Display Yokadi current version')


def processDataDirArg(dataDir):
    if dataDir:
        dataDir = os.path.abspath(dataDir)
        if not os.path.isdir(dataDir):
            tui.error("Directory '{}' does not exist".format(dataDir))
            sys.exit(1)
    else:
        dataDir = basepaths.getDataDir()
        os.makedirs(dataDir, exist_ok=True)
    return dataDir


def processDbPathArg(dbPath, dataDir):
    if not dbPath:
        return basepaths.getDbPath(dataDir)
    dbPath = os.path.abspath(dbPath)
    dbDir = os.path.dirname(dbPath)
    tui.warning('--db option is deprecated and will be removed in the next version, use --datadir instead')
    if not os.path.isdir(dbDir):
        tui.error("Directory '{}' does not exist".format(dbDir))
        sys.exit(1)
    return dbPath


def warnYokadiDbEnvVariable():
    if os.getenv('YOKADI_DB'):
        tui.warning('The YOKADI_DB environment variable is deprecated and will be removed in the next version, use the'
                    ' --datadir command-line option instead')


def processArgs(args):
    if args.version:
        print('Yokadi - {}'.format(yokadi.__version__))
        sys.exit(0)
    warnYokadiDbEnvVariable()
    dataDir = processDataDirArg(args.dataDir)
    dbPath = processDbPathArg(args.dbPath, dataDir)
    return dataDir, dbPath



# -*- coding: UTF-8 -*-
"""
Update from version 4 to version 5 of Yokadi DB

@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or newer
"""


def updateBugsKeywordsNames(cursor):
    for keyword in ("bug", "severity", "likelihood"):
        cursor.execute("update keyword set name='_%s' where name='%s'" % (keyword, keyword))


def removeTextWidthParam(cursor):
    cursor.execute("delete from config where name='TEXT_WIDTH'")


def update(cursor):
    updateBugsKeywordsNames(cursor)
    removeTextWidthParam(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Yokadi daemon. Used to monitor due tasks and warn user.

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""

import sys
import os
import time
from datetime import datetime, timedelta
from signal import SIGTERM, SIGHUP, signal
from subprocess import Popen
from argparse import ArgumentParser

from yokadi.core import fileutils

try:
    import setproctitle
except ImportError:
    print("You don't have the setproctitle package.")
    print("Get it on http://pypi.python.org/pypi/setproctitle/")
    print("Or use 'easy_install setproctitle'")
    sys.exit(1)

from yokadi.core.daemon import Daemon
from yokadi.core import basepaths
from yokadi.yical.yical import YokadiIcalServer

from yokadi.core import db
from yokadi.core.db import Project, Task, getConfigKey
from yokadi.ycli import commonargs


# Daemon polling delay (in seconds)
PROCESS_INTERVAL = 30
EVENTLOOP_INTERVAL = 1

# Ical daemon default port
DEFAULT_TCP_ICAL_PORT = 8000

# Event sender to main loop
event = [True, ""]


def sigTermHandler(signal, stack):
    """Handler when yokadid receive SIGTERM"""
    print("Receive SIGTERM. Exiting")
    print("End of yokadi Daemon")
    event[0] = False
    event[1] = "SIGTERM"


def sigHupHandler(signal, stack):
    """Handler when yokadid receive SIGHUP"""
    print("Receive SIGHUP. Reloading configuration")
    event[0] = False
    event[1] = "SIGHUP"


def eventLoop():
    """Main event loop"""
    delta = timedelta(hours=float(getConfigKey("ALARM_DELAY")))
    suspend = timedelta(hours=float(getConfigKey("ALARM_SUSPEND")))
    cmdDelayTemplate = getConfigKey("ALARM_DELAY_CMD")
    cmdDueTemplate = getConfigKey("ALARM_DUE_CMD")
    session = db.getSession()
    # For the two following dict, task id is key, and value is (duedate, triggerdate)
    triggeredDelayTasks = {}
    triggeredDueTasks = {}
    activeTaskFilter = [Task.status != "done",
                        Task.projectId == Project.id,
                        Project.active == True]  # noqa

    def process(now):
        delayTasks = session.query(Task).filter(Task.dueDate < now + delta,
                                                Task.dueDate > now,
                                                *activeTaskFilter)
        dueTasks = session.query(Task).filter(Task.dueDate < now,
                                              *activeTaskFilter)
        processTasks(delayTasks, triggeredDelayTasks, cmdDelayTemplate, suspend)
        processTasks(dueTasks, triggeredDueTasks, cmdDueTemplate, suspend)

    nextProcessTime = datetime.today().replace(microsecond=0)
    while event[0]:
        now = datetime.today().replace(microsecond=0)
        if now > nextProcessTime:
            process(now)
            nextProcessTime = now + timedelta(seconds=PROCESS_INTERVAL)
        time.sleep(EVENTLOOP_INTERVAL)


def processTasks(tasks, triggeredTasks, cmdTemplate, suspend):
    """Process a list of tasks and trigger action if needed
    @param tasks: list of tasks
    @param triggeredTasks: dict of tasks that has been triggered. Dict can be updated
    @param cmdTemplate: command line template to execute if task trigger
    @param suspend: timedelta beetween to task trigger"""
    now = datetime.now()
    for task in tasks:
        if task.id in triggeredTasks and triggeredTasks[task.id][0] == task.dueDate:
            # This task with the same dueDate has already been triggered
            if now - triggeredTasks[task.id][1] < suspend:
                # Task has been trigger recently, skip to next
                continue
        print("Task %s is due soon" % task.title)
        cmd = cmdTemplate.replace("{ID}", str(task.id))
        cmd = cmd.replace("{TITLE}", task.title.replace('"', '\"'))
        cmd = cmd.replace("{PROJECT}", task.project.name.replace('"', '\"'))
        cmd = cmd.replace("{DATE}", str(task.dueDate))
        process = Popen(cmd, shell=True)
        process.wait()
        # TODO: redirect stdout/stderr properly to Log (not so easy...)
        triggeredTasks[task.id] = (task.dueDate, datetime.now())


def killYokadid(pidFile):
    """Kill Yokadi daemon
    @param pidFile: file where the pid of the daemon is stored
    """
    # reuse Daemon.stop() code
    daemon = Daemon(pidFile)
    daemon.stop()


def parseOptions(defaultPidFile, defaultLogFile):
    parser = ArgumentParser()

    commonargs.addArgs(parser)

    parser.add_argument("-i", "--icalserver",
                        dest="icalserver", default=False, action="store_true",
                        help="Start the optional HTTP Ical Server")

    parser.add_argument("-p", "--port",
                        dest="tcpPort", default=DEFAULT_TCP_ICAL_PORT,
                        help="TCP port of ical server (default: %s)" % DEFAULT_TCP_ICAL_PORT,
                        metavar="PORT")

    parser.add_argument("-l", "--listen",
                        dest="tcpListen", default=False, action="store_true",
                        help="Listen on all interface (not only localhost) for ical server")

    parser.add_argument("-k", "--kill",
                        dest="kill", default=False, action="store_true",
                        help="Kill the Yokadi daemon. The daemon is found from the process ID stored in the file"
                             " specified with --pid")

    parser.add_argument("--restart",
                        dest="restart", default=False, action="store_true",
                        help="Restart the Yokadi daemon. The daemon is found from the process ID stored in the file"
                             " specified with --pid")

    parser.add_argument("-f", "--foreground",
                        dest="foreground", default=False, action="store_true",
                        help="Don't fork background. Useful for debug")

    parser.add_argument("--pid",
                        dest="pidFile", default=defaultPidFile,
                        help="File in which Yokadi daemon stores its process ID (default: %s)" % defaultPidFile)

    parser.add_argument("--log",
                        dest="logFile", default=defaultLogFile,
                        help="File in which Yokadi daemon stores its log output (default: %s)" % defaultLogFile)

    return parser.parse_args()


class YokadiDaemon(Daemon):
    def __init__(self, dbPath, options):
        Daemon.__init__(self, options.pidFile, stdout=options.logFile, stderr=options.logFile)
        self.dbPath = dbPath
        self.options = options

    def run(self):
        db.connectDatabase(self.dbPath, createIfNeeded=False)
        print("Using %s" % self.dbPath)
        session = db.getSession()

        # Basic tests :
        if not len(session.query(db.Config).all()) >= 1:
            print("Your database seems broken or not initialised properly. Start yokadi command line tool to do it")
            sys.exit(1)

        # Start ical http handler
        if self.options.icalserver:
            yokadiIcalServer = YokadiIcalServer(self.options.tcpPort, self.options.tcpListen)
            yokadiIcalServer.start()

        # Start the main event Loop
        try:
            while event[1] != "SIGTERM":
                eventLoop()
                event[0] = True
        except KeyboardInterrupt:
            print("\nExiting...")


def main():
    # TODO: check that yokadid is not already running for this database ? Not very harmful...
    # Set process name to "yokadid"
    setproctitle.setproctitle("yokadid")

    # Make the event list global to allow communication with main event loop
    global event

    defaultPidFile = os.path.join(basepaths.getRuntimeDir(), "yokadid.pid")
    defaultLogFile = os.path.join(basepaths.getLogDir(), "yokadid.log")
    args = parseOptions(defaultPidFile, defaultLogFile)
    _, dbPath = commonargs.processArgs(args)

    if args.kill:
        killYokadid(args.pidFile)
        sys.exit(0)

    if args.pidFile == defaultPidFile:
        fileutils.createParentDirs(args.pidFile, mode=0o700)

    if args.logFile == defaultLogFile:
        fileutils.createParentDirs(args.logFile, mode=0o700)

    signal(SIGTERM, sigTermHandler)
    signal(SIGHUP, sigHupHandler)

    if args.restart:
        daemon = YokadiDaemon(dbPath, args)
        daemon.restart()

    daemon = YokadiDaemon(dbPath, args)
    if args.foreground:
        daemon.run()
    else:
        daemon.start()


if __name__ == "__main__":
    main()



"""
Update from version 10 to version 11 of Yokadi DB

- Make the tuple (task_id, keyword_id) unique in TaskKeyword

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""
from collections import defaultdict

from yokadi.update import updateutils


def removeTaskKeywordDuplicates(cursor):
    sql = "select id, task_id, keyword_id from task_keyword"

    # Create a dict of (task_id, keyword_id) => [id...]
    dct = defaultdict(list)
    for row in cursor.execute(sql).fetchall():
        tk_id, task_id, keyword_id = row
        dct[(task_id, keyword_id)].append(tk_id)

    # Delete all extra ids
    for (task_id, keyword_id), tk_ids in dct.items():
        for tk_id in tk_ids[1:]:
            cursor.execute("delete from task_keyword where id = ?", (tk_id,))


def update(cursor):
    removeTaskKeywordDuplicates(cursor)


if __name__ == "__main__":
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Bug test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.ycli import tui
from yokadi.ycli.main import YokadiCmd
from yokadi.core import db, dbutils
from yokadi.core.db import Task, setDefaultConfig
from yokadi.core.yokadiexception import YokadiException


class BugTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        setDefaultConfig()
        tui.clearInputAnswers()
        self.cmd = YokadiCmd()

    def testAdd(self):
        tui.addInputAnswers("y", "2", "4", "123")
        self.cmd.do_bug_add("x t1")

        tui.addInputAnswers("n")
        self.cmd.do_bug_add("notExistingProject newBug")

        tasks = self.session.query(Task).all()
        result = [x.title for x in tasks]
        expected = ["t1"]
        self.assertEqual(result, expected)

        kwDict = self.session.get(Task, 1).getKeywordDict()
        self.assertEqual(kwDict, dict(_severity=2, _likelihood=4, _bug=123))

        for bad_input in ("",  # No project
                          "x"):  # No task name
            self.assertRaises(YokadiException, self.cmd.do_bug_add, bad_input)

    def testEdit(self):
        task = dbutils.addTask("prj", "bug", interactive=False)
        kwDict = dict(_severity=1, _likelihood=2, _bug=3)
        task.setKeywordDict(kwDict)
        self.session.commit()

        tui.addInputAnswers("bug edited", "2", "4", "6")
        self.cmd.do_bug_edit(str(task.id))

        task = dbutils.getTaskFromId(task.id)
        self.assertEqual(task.title, "bug edited")
        kwDict = task.getKeywordDict()
        self.assertEqual(kwDict, dict(_severity=2, _likelihood=4, _bug=6))

# vi: ts=4 sw=4 et



#!/usr/bin/env python3
"""
Yokadi unit tests

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""

import unittest
import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), os.pardir, os.pardir))

try:
    import icalendar  # noqa: F401
    hasIcalendar = True
except ImportError:
    hasIcalendar = False
    print("icalendar is not installed, some tests won't be run")

from parseutilstestcase import ParseUtilsTestCase  # noqa: F401, E402
from yokadioptionparsertestcase import YokadiOptionParserTestCase  # noqa: F401, E402
from ydateutilstestcase import YDateUtilsTestCase  # noqa: F401, E402
from dbutilstestcase import DbUtilsTestCase  # noqa: F401, E402
from projecttestcase import ProjectTestCase  # noqa: F401, E402
from completerstestcase import CompletersTestCase  # noqa: F401, E402
from tasktestcase import TaskTestCase  # noqa: F401, E402
from bugtestcase import BugTestCase  # noqa: F401, E402
from aliastestcase import AliasTestCase  # noqa: F401, E402
from textlistrenderertestcase import TextListRendererTestCase  # noqa: F401, E402
if hasIcalendar:
    from icaltestcase import IcalTestCase  # noqa: F401, E402
from keywordtestcase import KeywordTestCase  # noqa: F401, E402
from tuitestcase import TuiTestCase  # noqa: F401, E402
from helptestcase import HelpTestCase  # noqa: F401, E402
from conftestcase import ConfTestCase  # noqa: F401, E402
from massedittestcase import MassEditTestCase  # noqa: F401, E402
from basepathstestcase import BasePathsUnixTestCase, BasePathsWindowsTestCase  # noqa: F401, E402
from keywordfiltertestcase import KeywordFilterTestCase  # noqa: F401, E402
from recurrenceruletestcase import RecurrenceRuleTestCase  # noqa: F401, E402
from argstestcase import ArgsTestCase  # noqa: F401, E402
from dbtestcase import DbTestCase  # noqa: F401, E402


def main():
    unittest.main()


if __name__ == "__main__":
    main()
# vi: ts=4 sw=4 et



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Command line oriented, sqlite powered, todo list

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""

import locale
import os
import platform
import sys

import readline

import traceback
from cmd import Cmd
from argparse import ArgumentParser

import sqlalchemy
from colorama import just_fix_windows_console

import yokadi

from yokadi.core import db
from yokadi.core import basepaths
from yokadi.core import fileutils
from yokadi.update import update

from yokadi.ycli import tui, commonargs
from yokadi.ycli.aliascmd import AliasCmd, resolveAlias
from yokadi.ycli.confcmd import ConfCmd
from yokadi.ycli.keywordcmd import KeywordCmd
from yokadi.ycli.projectcmd import ProjectCmd
from yokadi.ycli.taskcmd import TaskCmd
from yokadi.core.yokadiexception import YokadiException, BadUsageException
from yokadi.core.yokadioptionparser import YokadiOptionParserNormalExitException

readline.parse_and_bind("set show-all-if-ambiguous on")


# TODO: move YokadiCmd to a separate module in ycli package
class YokadiCmd(TaskCmd, ProjectCmd, KeywordCmd, ConfCmd, AliasCmd, Cmd):
    def __init__(self):
        Cmd.__init__(self)
        TaskCmd.__init__(self)
        ProjectCmd.__init__(self)
        KeywordCmd.__init__(self)
        AliasCmd.__init__(self)
        ConfCmd.__init__(self)
        self.prompt = "yokadi> "
        self.historyPath = basepaths.getHistoryPath()
        self.loadHistory()

    def emptyline(self):
        """Executed when input is empty. Reimplemented to do nothing."""
        return

    def default(self, line):
        nline = resolveAlias(line, self.aliases)
        if nline != line:
            return self.onecmd(nline)
        elif nline.isdigit():
            self.do_t_show(nline)
        elif nline == "_":
            self.do_t_show(nline)
        else:
            raise YokadiException("Unknown command. Use 'help' to see all available commands")

    def completedefault(self, text, line, begidx, endidx):
        """Default completion command.
        Try to see if command is an alias and find the
        appropriate complete function if it exists"""
        nline = resolveAlias(line, self.aliases)
        compfunc = getattr(self, 'complete_' + nline.split()[0])
        matches = compfunc(text, line, begidx, endidx)
        return matches

    def do_EOF(self, line):
        """Quit."""
        print()
        return True

    # Some standard alias
    do_quit = do_EOF
    do_q = do_EOF
    do_exit = do_EOF

    def onecmd(self, line):
        """This method is subclassed just to be
        able to encapsulate it with a try/except bloc"""
        try:
            # Decode user input
            line = line
            return Cmd.onecmd(self, line)
        except YokadiOptionParserNormalExitException:
            pass
        except UnicodeDecodeError as e:
            tui.error("Unicode decoding error. Please check you locale and terminal settings (%s)." % e)
        except UnicodeEncodeError as e:
            tui.error("Unicode encoding error. Please check you locale and terminal settings (%s)." % e)
        except BadUsageException as e:
            tui.error("*** Bad usage ***\n\t%s" % e)
            cmd = line.split(' ')[0]
            self.do_help(cmd)
        except YokadiException as e:
            tui.error("*** Yokadi error ***\n\t%s" % e)
        except IOError as e:
            # We can get I/O errors when yokadi is piped onto another shell commands
            # that breaks.
            print("*** I/O error ***\n\t%s" % e, file=sys.stderr)
        except KeyboardInterrupt:
            print("*** Break ***")
        except Exception as e:
            tui.error("Unhandled exception (oups)\n\t%s" % e)
            print("This is a bug of Yokadi, sorry.")
            print("Send the above message by email to Yokadi developers (ml-yokadi@sequanux.org) to help them make"
                  " Yokadi better.")
            cut = "---------------------8<----------------------------------------------"
            print(cut)
            traceback.print_exc()
            print("--")
            print("Python: %s" % sys.version.replace("\n", " "))
            print("SQL Alchemy: %s" % sqlalchemy.__version__)
            print("OS: %s" % platform.system())
            print("Yokadi: %s" % yokadi.__version__)
            print(cut)
            print()

    def loadHistory(self):
        """Tries to load previous history list from disk"""
        try:
            readline.read_history_file(self.historyPath)
        except Exception:
            # Cannot load any previous history. Start from a clear one
            pass

    def writeHistory(self):
        """Writes shell history to disk"""
        try:
            fileutils.createParentDirs(self.historyPath)
            # Open r/w and close file to create one if needed
            historyFile = open(self.historyPath, "w", encoding='utf-8')
            historyFile.close()
            readline.set_history_length(1000)
            readline.write_history_file(self.historyPath)
        except Exception as e:
            tui.warning("Fail to save history to %s. Error was:\n\t%s"
                        % (self.historyPath, e))

    def do_help(self, arg):
        """Type help <topic> to see the help for a given topic"""

        """
        Overload do_help to show help from the command parser if it exists:
        if there is a parser_foo() method, assume this method returns a
        YokadiOptionParser for the do_foo() method and show the help of the
        parser, instead of do_foo() docstring.
        """
        if arg in self.aliases:
            # If arg is an alias, get help on the underlying command
            arg = self.aliases[arg].split()[0]
        if hasattr(self, "parser_" + arg):
            parserMethod = getattr(self, "parser_" + arg)
            parserMethod().print_help(sys.stderr)
        else:
            print("Usage: ", end=' ')
            Cmd.do_help(self, arg)

    def completenames(self, text, *ignored):
        """Complete commands names. Same as Cmd.cmd one but with support
        for command aliases. Code kindly borrowed to Pysql"""
        dotext = 'do_' + text
        names = [a[3:] for a in self.get_names() if a.startswith(dotext)]
        names.extend([a for a in list(self.aliases.keys()) if a.startswith(text)])
        return names


def createArgumentParser():
    parser = ArgumentParser()
    commonargs.addArgs(parser)

    parser.add_argument("-c", "--create-only",
                        dest="createOnly", default=False, action="store_true",
                        help="Just create an empty database")

    parser.add_argument("-u", "--update",
                        dest="update", action="store_true",
                        help="Update database to the latest version")

    parser.add_argument('cmd', nargs='*')
    return parser


def main():
    locale.setlocale(locale.LC_ALL, os.environ.get("LANG", "C"))
    just_fix_windows_console()

    parser = createArgumentParser()
    args = parser.parse_args()
    dataDir, dbPath = commonargs.processArgs(args)

    basepaths.migrateOldHistory()
    try:
        basepaths.migrateOldDb(dbPath)
    except basepaths.MigrationException as exc:
        print(exc)
        return 1

    if args.update:
        return update.update(dbPath)

    try:
        db.connectDatabase(dbPath)
    except db.DbUserException as exc:
        print(exc)
        return 1

    if args.createOnly:
        return 0
    db.setDefaultConfig()  # Set default config parameters

    cmd = YokadiCmd()

    try:
        if len(args.cmd) > 0:
            print(" ".join(args.cmd))
            cmd.onecmd(" ".join(args.cmd))
        else:
            cmd.cmdloop()
    except KeyboardInterrupt:
        print("\n\tBreak ! (the nice way to quit is 'quit' or 'EOF' (ctrl-d)")
        return 1
    # Save history
    cmd.writeHistory()
    return 0


if __name__ == "__main__":
    sys.exit(main())
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
An ArgumentParser which accepts a single string as input and raise an exception
instead of calling sys.exit() in case of error

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from argparse import ArgumentParser
import sys

from yokadi.core.yokadiexception import YokadiException


class YokadiOptionParserNormalExitException(YokadiException):
    """A dummy exception which makes it possible to have --help exit silently"""
    pass


class YokadiOptionParser(ArgumentParser):
    def __init__(self, prog=None):
        ArgumentParser.__init__(self, prog=prog)

    def parse_args(self, line):
        argv = line.split(" ")
        # Splitting an empty line gives us [""], not an empty array
        if argv == [""]:
            argv = []

        # Unknown options will throw an error
        args = ArgumentParser.parse_args(self, argv)
        return args

    def exit(self, status=0, msg=None):
        if msg:
            sys.stderr.write(msg)
        if status == 0:
            raise YokadiOptionParserNormalExitException()
        else:
            raise YokadiException(msg)

    def error(self, msg):
        self.print_usage(sys.stderr)
        raise YokadiException(msg)
# vi: ts=4 sw=4 et



"""
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.core import db


class DbTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()

    def testSetVersion(self):
        newVersion = db.DB_VERSION + 1
        db._database.setVersion(newVersion)
        version = db._database.getVersion()
        self.assertEqual(version, newVersion)



# -*- coding: UTF-8 -*-
"""
Update from version 7 to version 8 of Yokadi DB
Drops the projectkeyword table, since we are removing this feature.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""


def dropProjectKeywordTable(cursor):
    cursor.execute('drop table project_keyword')


def update(cursor):
    dropProjectKeywordTable(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Project test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.core import db
from yokadi.core.db import Project, setDefaultConfig
from yokadi.ycli import completers


class CompletersTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        setDefaultConfig()
        self.session = db.getSession()

    def testProjectCompleter(self):
        self.session.add_all([Project(name="foo"),
                             Project(name="foo2"),
                             Project(name="bar")])

        expected = ["foo ", "foo2 "]
        completer = completers.ProjectCompleter(1)
        result = completer("f", "t_add f", 6, 8)
        self.assertEqual(result, expected)

    def testCompleteParameterPosition(self):
        data = [
            (("bla", "t_add bla", 6, 10), 1),
            (("bli", "t_add bla bli", 10, 14), 2),
        ]
        for params, expectedResult in data:
            result = completers.computeCompleteParameterPosition(*params)
            self.assertEqual(result, expectedResult)
# vi: ts=4 sw=4 et



"""
Utilities to handle database schema updates

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""
import sqlite3
import sys


class UpdateError(Exception):
    pass


class UpdateCanceledError(UpdateError):
    def __init__(self):
        super(UpdateError, self).__init__("Canceled")


def getTableList(cursor):
    cursor.execute("select name from sqlite_master where type='table' and name!='sqlite_sequence'")
    return [x[0] for x in cursor.fetchall()]


def getTableColumnList(cursor, table):
    cursor.execute("select * from %s" % table)
    return [x[0] for x in cursor.description]


def deleteTableColumns(cursor, table, columnsToDelete):
    columnList = getTableColumnList(cursor, table)
    for column in columnsToDelete:
        columnList.remove(column)
    columns = ",".join(columnList)
    sqlCommands = (
        "create temporary table {table}_backup({columns})",
        "insert into {table}_backup select {columns} from {table}",
        "drop table {table}",
        "create table {table}({columns})",
        "insert into {table} select {columns} from {table}_backup",
        "drop table {table}_backup",
    )
    for sql in sqlCommands:
        cursor.execute(sql.format(table=table, columns=columns))


def main(function):
    """
    Runs an update function on a database. Useful as a test main
    """
    dbpath = sys.argv[1]
    with sqlite3.connect(dbpath) as conn:
        function(conn.cursor())



"""
Parse utilities which do not need the db module.

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import re
from yokadi.core.yokadiexception import YokadiException


gSimplifySpaces = re.compile("  +")


def simplifySpaces(line):
    line = gSimplifySpaces.subn(" ", line)[0]
    line = line.strip()
    return line


def parseParameters(line):
    """Parse line of form -a -b -c some text
    @return: ((a, b, c), some text)
    """
    parameters = []
    text = []
    line = simplifySpaces(line)
    for word in line.split():
        if word.startswith("-") and len(word) == 2:
            parameters.append(word[1])
        else:
            text.append(word)
    return (parameters, " ".join(text))


def parseOneWordName(line):
    """Parse line, check it is a one word project name and return it
    @return: the name
    """
    line = line.strip()
    if " " in line:
        raise YokadiException("Name cannot contain spaces")
    if not line:
        raise YokadiException("Name cannot be empty")
    return line



"""
Command line argument test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os

from argparse import ArgumentParser
from tempfile import TemporaryDirectory

from yokadi.tests.yokaditestcase import YokadiTestCase

from yokadi.core import basepaths
from yokadi.ycli import commonargs


def parseArgs(argv):
    parser = ArgumentParser()
    commonargs.addArgs(parser)
    return parser.parse_args(argv)


class ArgsTestCase(YokadiTestCase):
    def setUp(self):
        super().setUp()
        self.defaultDataDir = basepaths.getDataDir()
        self.defaultDbPath = basepaths.getDbPath(self.defaultDataDir)

    def testNoArguments(self):
        args = parseArgs([])
        dataDir, dbPath = commonargs.processArgs(args)
        self.assertEqual(dataDir, self.defaultDataDir)
        self.assertEqual(dbPath, self.defaultDbPath)

        self.assertTrue(os.path.isdir(dataDir))
        self.assertTrue(os.path.isdir(os.path.dirname(dbPath)))

    def testDataDir(self):
        with TemporaryDirectory(prefix="yokadi-tests-") as tmpDir:
            args = parseArgs(["--datadir", tmpDir])
            dataDir, dbPath = commonargs.processArgs(args)
            self.assertEqual(dataDir, tmpDir)
            self.assertEqual(dbPath, os.path.join(tmpDir, basepaths.DB_NAME))

    def testRelativeDataDir(self):
        with TemporaryDirectory(prefix="yokadi-tests-") as tmpDir:
            os.chdir(tmpDir)
            args = parseArgs(["--datadir", "."])
            dataDir, dbPath = commonargs.processArgs(args)
            self.assertEqual(dataDir, tmpDir)
            self.assertEqual(dbPath, os.path.join(tmpDir, basepaths.DB_NAME))

    def testDataDirDoesNotExist(self):
        args = parseArgs(["--datadir", "/does/not/exist"])
        self.assertRaises(SystemExit, commonargs.processArgs, args)

    def testCantUseBothDataDirAndDb(self):
        self.assertRaises(SystemExit, parseArgs, ["--datadir", "foo", "--db", "bar"])

    def testDb(self):
        with TemporaryDirectory(prefix="yokadi-tests-") as tmpDir:
            args = parseArgs(["--db", os.path.join(tmpDir, "foo.db")])
            dataDir, dbPath = commonargs.processArgs(args)
            self.assertEqual(dataDir, self.defaultDataDir)
            self.assertEqual(dbPath, os.path.join(tmpDir, "foo.db"))

    def testRelativeDb(self):
        with TemporaryDirectory(prefix="yokadi-tests-") as tmpDir:
            os.chdir(tmpDir)
            args = parseArgs(["--db", "foo.db"])
            dataDir, dbPath = commonargs.processArgs(args)
            self.assertEqual(dataDir, self.defaultDataDir)
            self.assertEqual(dbPath, os.path.join(tmpDir, "foo.db"))

    def testDbDirDoesNotExist(self):
        args = parseArgs(["--db", "/does/not/exist/foo.db"])
        self.assertRaises(SystemExit, commonargs.processArgs, args)

    def testArgsOverrideEnvVar(self):
        with TemporaryDirectory(prefix="yokadi-tests-") as tmpDir:
            os.environ["YOKADI_DB"] = os.path.join(tmpDir, "env.db")
            os.chdir(tmpDir)
            args = parseArgs(["--db", "arg.db"])
            dataDir, dbPath = commonargs.processArgs(args)
            self.assertEqual(dataDir, self.defaultDataDir)
            self.assertEqual(dbPath, os.path.join(tmpDir, "arg.db"))



# -*- coding: utf-8 -*-
"""
Yokadi main package
@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license:GPL v3 or later
"""

__version__ = "1.3.0"



# -*- coding: UTF-8 -*-
"""
Csv rendering of t_list output

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import csv

TASK_FIELDS = ["title", "creationDate", "dueDate", "doneDate", "description",
               "urgency", "status", "project", "keywords"]


class CsvListRenderer(object):
    def __init__(self, out):
        self.writer = csv.writer(out, dialect="excel")
        self._writerow(TASK_FIELDS)  # Header

    def addTaskList(self, project, taskList):
        for task in taskList:
            row = [getattr(task, field) for field in TASK_FIELDS if field != "keywords"]
            row.append(task.getKeywordsAsString())
            self._writerow(row)

    def end(self):
        pass

    def _writerow(self, row):
        self.writer.writerow([str(x) for x in row])
# vi: ts=4 sw=4 et



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Command line oriented, sqlite powered, todo list

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""
import os
import sys
from argparse import ArgumentParser

from yokadi.core import db
from yokadi.core import dbutils
from yokadi.core import ydateutils

PROJECTS = ["birthday", "work", "home"]

KEYWORDS = ["phone", "grocery", "_note"]


def main():
    parser = ArgumentParser()
    parser.add_argument('db', metavar='<db>')
    args = parser.parse_args()

    dbname = args.db
    if os.path.exists(dbname):
        os.unlink(dbname)

    db.connectDatabase(dbname)
    db.setDefaultConfig()
    session = db.getSession()

    for name in PROJECTS:
        session.add(db.Project(name=name))

    for name in KEYWORDS:
        dbutils.getOrCreateKeyword(name, interactive=False)

    dbutils.addTask("birthday", "Buy food", {"grocery": None})
    dbutils.addTask("birthday", "Buy drinks", {"grocery": None})
    dbutils.addTask("birthday", "Invite Bob", {"phone": None})
    dbutils.addTask("birthday", "Invite Wendy", {"phone": None})
    dbutils.addTask("birthday", "Bake a yummy cake")
    dbutils.addTask("birthday", "Decorate living-room")

    task = dbutils.addTask("home", "Fix leak in the roof")
    task.dueDate = ydateutils.parseHumaneDateTime("-2d")

    dbutils.addTask("home", "Buy AAA batteries for kid toys", {"grocery": None})

    task = dbutils.addTask("home", "Bring the car to the garage")
    task.dueDate = ydateutils.parseHumaneDateTime("-1d")
    task.status = "done"

    task = dbutils.addTask("work", "Finish weekly report")
    task.dueDate = ydateutils.parseHumaneDateTime("+4d")
    task.description = """Include results from Acme department: http://acme.intranet/results.
    Don't forget to CC boss@acme.intranet.
    """

    session.commit()

    return 0


if __name__ == "__main__":
    sys.exit(main())
# vi: ts=4 sw=4 et



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
This script updates a Yokadi database to the latest version

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""

import os
import shutil
import sqlite3
import sys
import time
from argparse import ArgumentParser
from tempfile import TemporaryDirectory

from yokadi.core import db
from yokadi.update import updateutils

# Those modules look unused, but they are used "dynamically"
from yokadi.update import update1to2  # noqa
from yokadi.update import update2to3  # noqa
from yokadi.update import update3to4  # noqa
from yokadi.update import update4to5  # noqa
from yokadi.update import update5to6  # noqa
from yokadi.update import update6to7  # noqa
from yokadi.update import update7to8  # noqa
from yokadi.update import update8to9  # noqa
from yokadi.update import update9to10  # noqa
from yokadi.update import update10to11  # noqa
from yokadi.update import update11to12  # noqa


def getVersion(fileName):
    database = db.Database(fileName, createIfNeeded=False, updateMode=True)
    return database.getVersion()


def setVersion(fileName, version):
    database = db.Database(fileName, createIfNeeded=False, updateMode=True)
    database.setVersion(version)


def importTable(dstCursor, srcCursor, table):
    columns = updateutils.getTableColumnList(dstCursor, table)
    columnString = ", ".join(columns)
    sql = "select {} from {}".format(columnString, table)

    placeHolders = ", ".join(["?"] * len(columns))
    insertSql = "insert into {}({}) values({})".format(table, columnString, placeHolders)

    query = srcCursor.execute(sql)
    while True:
        rows = query.fetchmany(size=100)
        if not rows:
            break
        dstCursor.executemany(insertSql, rows)


def recreateDb(workPath, destPath):
    assert os.path.exists(workPath)

    print("Recreating the database")
    database = db.Database(destPath, createIfNeeded=True, updateMode=True)  # noqa

    print("Importing content to the new database")
    srcConn = sqlite3.connect(workPath)
    srcCursor = srcConn.cursor()
    dstConn = sqlite3.connect(destPath)
    dstCursor = dstConn.cursor()

    for table in updateutils.getTableList(dstCursor):
        importTable(dstCursor, srcCursor, table)
    dstConn.commit()


def err(message):
    print("error: " + message, file=sys.stderr)


def update(dbPath, newDbPath=None, inplace=True):
    # Check paths
    if not os.path.exists(dbPath):
        err("'{}' does not exist.".format(dbPath))
        return 1

    if not inplace and os.path.exists(newDbPath):
        err("'{}' already exists.".format(newDbPath))
        return 1

    # Check version
    version = getVersion(dbPath)
    print("Found version %d" % version)

    if version == db.DB_VERSION:
        print("Nothing to do")
        return 0

    if inplace:
        destDir = os.path.dirname(dbPath)
    else:
        destDir = os.path.dirname(newDbPath)

    with TemporaryDirectory(prefix="yokadi-update-", dir=destDir) as tempDir:
        # Copy the DB
        workDbPath = os.path.join(tempDir, "work.db")
        shutil.copy(dbPath, workDbPath)

        # Start import
        oldVersion = getVersion(workDbPath)

        with sqlite3.connect(workDbPath) as conn:
            cursor = conn.cursor()

            for version in range(oldVersion, db.DB_VERSION):
                moduleName = "update{}to{}".format(version, version + 1)
                print("Updating to {}".format(version + 1))
                function = globals()[moduleName].update
                function(cursor)

        setVersion(workDbPath, db.DB_VERSION)

        # Recreate the DB
        recreatedDbPath = os.path.join(tempDir, "recreated.db")
        recreateDb(workDbPath, recreatedDbPath)

        # Move to final paths
        if inplace:
            base, ext = os.path.splitext(dbPath)
            timestamp = time.strftime("%Y%m%d")
            backupPath = base + "-v{}-{}".format(oldVersion, timestamp) + ext
            os.rename(dbPath, backupPath)
            print("Old database renamed to {}".format(backupPath))
            os.rename(recreatedDbPath, dbPath)
        else:
            os.rename(recreatedDbPath, newDbPath)

    return 0


def main():
    # Parse args
    parser = ArgumentParser()
    parser.add_argument('current', metavar='<path/to/current.db>',
                        help="Path to the database to update.")
    parser.add_argument('updated', metavar='<path/to/updated.db>',
                        help="Path to the destination database. Mandatory unless --inplace is used",
                        nargs="?")
    parser.add_argument("-i", "--in-place",
                        dest="inplace", action="store_true",
                        help="Replace current file")

    args = parser.parse_args()

    dbPath = os.path.abspath(args.current)

    if args.inplace:
        newDbPath = None
    else:
        newDbPath = os.path.abspath(args.updated)

    try:
        return update(dbPath, newDbPath, inplace=args.inplace)
    except updateutils.UpdateError as exc:
        err(str(exc))
        return 1


if __name__ == "__main__":
    sys.exit(main())
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Parse utilities. Used to manipulate command line text.

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
from yokadi.ycli import tui
from yokadi.ycli.basicparseutils import simplifySpaces
from yokadi.core import db
from yokadi.core.db import Keyword
from yokadi.core.dbutils import KeywordFilter
from yokadi.core.yokadiexception import YokadiException


def parseLine(line):
    """Parse line of form:
    project some text @keyword1 @keyword2=12 some other text
    @return: a tuple of ("project", "some text some other text", keywordDict"""

    # First extract project name
    line = simplifySpaces(line)
    if line.count(" "):
        project, line = line.split(" ", 1)
    else:
        project = line
        line = ""

    line, keywordFilters = extractKeywords(line)

    return project, line, keywordFiltersToDict(keywordFilters)


def extractKeywords(line):
    """Extract keywords (@k1 @k2=n..) from line
    @param line: line from which keywords are extracted
    @returns: (remaining_text, keywordFilters)"""
    keywordFilters = []
    remainingText = []
    for token in line.split():
        if token.startswith("@") or token.startswith("!@"):
            keywordFilters.append(parseKeyword(token))
        else:
            remainingText.append(token)

    return (" ".join(remainingText), keywordFilters)


def createLine(projectName, title, keywordDict):
    tokens = []
    for keywordName, value in list(keywordDict.items()):
        if value:
            tokens.append("@" + keywordName + "=" + str(value))
        else:
            tokens.append("@" + keywordName)

    if projectName:
        tokens.insert(0, projectName)

    tokens.append(title)
    return " ".join(tokens)


def keywordFiltersToDict(keywordFilters):
    """Convert a list of KeywordFilter instnance to a simple keyword dictionary"""
    keywordDict = {}
    for keywordFilter in keywordFilters:
        keywordDict[keywordFilter.name] = keywordFilter.value
    return keywordDict


def warnIfKeywordDoesNotExist(keywordFilters):
    """Warn user is keyword does not exist
    @return: True if at least one keyword does not exist, else False"""
    session = db.getSession()
    doesNotExist = False
    for keyword in [k.name for k in keywordFilters]:
        if session.query(Keyword).filter(Keyword.name.like(keyword)).count() == 0:
            tui.error("Keyword %s is unknown." % keyword)
            doesNotExist = True
    return doesNotExist


def parseKeyword(line):
    """Parse given line to create a keyword filter
    @return: a KeywordFilter instance"""
    operators = ("!=", "=")
    if " " in line:
        raise YokadiException("Keyword filter should not contain spaces")

    name = None
    negative = False
    value = None
    valueOperator = None

    if line.startswith("!"):
        negative = True
        line = line[1:]

    if not line.startswith("@"):
        raise YokadiException("Keyword name must be prefixed with a @")

    line = line[1:]  # Squash @
    line = line.replace("==", "=")  # Tolerate == syntax
    for operator in operators:
        if operator in line:
            name, value = line.split(operator, 1)
            valueOperator = operator
            try:
                value = int(value)
            except ValueError:
                raise YokadiException("Value of %s keyword must be an integer (got %s)" %
                                      (name, value))
            break
    else:
        # No operator found, only keyword name has been provided
        name = line

    return KeywordFilter(name, negative=negative, value=value, valueOperator=valueOperator)

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Utils for unit-test
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from collections import OrderedDict

import os


def multiLinesAssertEqual(test, str1, str2):
    lst1 = str1.splitlines()
    lst2 = str2.splitlines()
    for row, lines in enumerate(zip(lst1, lst2)):
        line1, line2 = lines
        test.assertEqual(line1, line2, "Error line %d:\n%r\n!=\n%r" % (row + 1, line1, line2))
    test.assertEqual(len(lst1), len(lst2))


def assertQueryEmpty(test, query):
    lst = list(query)
    test.assertEqual(lst, [])


class TestRenderer(object):
    """
    A fake renderer, which stores all rendered tasks in:
    - taskDict: a dict for each section
    - tasks: a list of all tasks
    """
    def __init__(self):
        self.taskDict = OrderedDict()
        self.tasks = []

    def addTaskList(self, sectionName, taskList):
        self.taskDict[sectionName] = taskList
        self.tasks.extend(taskList)

    def end(self):
        pass


class EnvironSaver(object):
    """
    This class saves and restore the environment.

    Can be used manually or as a context manager.
    """
    def __init__(self):
        self.save()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.restore()
        return False

    def save(self):
        self.oldEnv = dict(os.environ)

    def restore(self):
        # Do not use `os.environ = env`: this would replace the special os.environ
        # object with a plain dict. We must update the *existing* object.
        os.environ.clear()
        os.environ.update(self.oldEnv)


# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Bug related commands.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from yokadi.ycli import tui

SEVERITY_PROPERTY_NAME = "_severity"
LIKELIHOOD_PROPERTY_NAME = "_likelihood"
BUG_PROPERTY_NAME = "_bug"
PROPERTY_NAMES = SEVERITY_PROPERTY_NAME, LIKELIHOOD_PROPERTY_NAME, BUG_PROPERTY_NAME

SEVERITY_LIST = [
    (1, "Documentation"),
    (2, "Localization"),
    (3, "Aesthetic issues"),
    (4, "Balancing: Enables degenerate usage strategies that harm the experience"),
    (5, "Minor usability: Impairs usability in secondary scenarios"),
    (6, "Major usability: Impairs usability in key scenarios"),
    (7, "Crash: Bug causes crash or data loss. Asserts in the Debug release"),
]

LIKELIHOOD_LIST = [
    (1, "Will affect almost no one"),
    (2, "Will only affect a few users"),
    (3, "Will affect average number of users"),
    (4, "Will affect most users"),
    (5, "Will affect all users"),
]


def computeUrgency(keywordDict):
    likelihood = keywordDict[LIKELIHOOD_PROPERTY_NAME]
    severity = keywordDict[SEVERITY_PROPERTY_NAME]
    maxUrgency = LIKELIHOOD_LIST[-1][0] * SEVERITY_LIST[-1][0]
    return int(100 * likelihood * severity / maxUrgency)


def editBugKeywords(keywordDict):
    severity = keywordDict.get(SEVERITY_PROPERTY_NAME, None)
    likelihood = keywordDict.get(LIKELIHOOD_PROPERTY_NAME, None)
    bug = keywordDict.get(BUG_PROPERTY_NAME, None)

    severity = tui.selectFromList(SEVERITY_LIST, prompt="Severity", default=severity)
    likelihood = tui.selectFromList(LIKELIHOOD_LIST, prompt="Likelihood", default=likelihood)
    bug = tui.enterInt(prompt="bug", default=bug)

    keywordDict[BUG_PROPERTY_NAME] = bug

    if severity:
        keywordDict[SEVERITY_PROPERTY_NAME] = severity

    if likelihood:
        keywordDict[LIKELIHOOD_PROPERTY_NAME] = likelihood
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Keyword filter test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.core import db, dbutils
from yokadi.core.dbutils import KeywordFilter


class KeywordFilterTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()

    def testSimpleFilter(self):
        t1 = dbutils.addTask("p1", "t1", interactive=False)
        t2 = dbutils.addTask("p1", "t2", keywordDict={"k1": None}, interactive=False)
        t3 = dbutils.addTask("p1", "t3", keywordDict={"k2": None}, interactive=False)
        t4 = dbutils.addTask("p1", "t4", keywordDict={"k1": None, "k2": None}, interactive=False)

        testData = [
            (KeywordFilter("k1"), {t2, t4}),
            (KeywordFilter("k%"), {t2, t3, t4}),
            (KeywordFilter("k1", negative=True), {t1, t3}),
            (KeywordFilter("k%", negative=True), {t1}),
        ]

        for flt, expected in testData:
            query = self.session.query(db.Task)
            query = flt.apply(query)
            resultSet = {x.title for x in query}
            expectedSet = {x.title for x in expected}
            self.assertEqual(resultSet, expectedSet)



# -*- coding: UTF-8 -*-
"""
Alias test cases
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""
import unittest

from contextlib import redirect_stdout
from io import StringIO


from yokadi.core import db
from yokadi.core.db import Alias
from yokadi.ycli.aliascmd import AliasCmd
from yokadi.ycli import colors
from yokadi.ycli import tui


class AliasTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        self.cmd = AliasCmd()

    def testList(self):
        self.cmd.do_a_add("b t_add")
        self.cmd.do_a_add("a t_list")
        out = StringIO()
        with redirect_stdout(out):
            self.cmd.do_a_list("")
            content = out.getvalue()
            self.assertEqual(content,
                             colors.BOLD + "a".ljust(10) + colors.RESET + "=> t_list\n"
                             + colors.BOLD + "b".ljust(10) + colors.RESET + "=> t_add\n")

    def testList_empty(self):
        out = StringIO()
        with redirect_stdout(out):
            self.cmd.do_a_list("")
            content = out.getvalue()
            self.assertTrue("No alias defined" in content)

    def testAdd(self):
        self.cmd.do_a_add("l t_list")
        self.cmd.do_a_add("la t_list -a")
        aliases = Alias.getAsDict(self.session)
        self.assertEqual(aliases["l"], "t_list")
        self.cmd.do_a_remove("l")
        self.cmd.do_a_remove("la")
        self.cmd.do_a_remove("unknown")
        aliases = Alias.getAsDict(self.session)
        self.assertEqual(aliases, {})

    def testEditName(self):
        self.cmd.do_a_add("l t_list")

        tui.addInputAnswers("ls")
        self.cmd.do_a_edit_name("l")

        aliases = Alias.getAsDict(self.session)
        self.assertEqual(aliases["ls"], "t_list")

    def testEditCommand(self):
        self.cmd.do_a_add("l t_list")

        tui.addInputAnswers("foo")
        self.cmd.do_a_edit_command("l")

        aliases = Alias.getAsDict(self.session)
        self.assertEqual(aliases["l"], "foo")



# -*- coding: UTF-8 -*-
"""
Update from version 6 to version 7 of Yokadi DB
This one is empty, database schema don't change
but we want to prevent use of SQLObject against a database created with SQLAlchemy
because SQLObject cannot read timestamps written by SQLAlchemy

@author: Benjamin Port <contact@benjaminport.fr>
@license: GPL v3 or newer
"""


def update(dbpath):
    pass


if __name__ == "__main__":
    pass
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
TUI module test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import os
import unittest

from yokadi.ycli import tui


class TuiTestCase(unittest.TestCase):
    def setUp(self):
        tui.clearInputAnswers()

    def testEditEmptyText(self):
        # Add an unused input answer to pass the isInteractive() check in editText()
        tui.addInputAnswers("")

        os.environ["EDITOR"] = "/usr/bin/true"
        out = tui.editText(None)
        self.assertEqual(out, "")

    def testEnterInt(self):
        tui.addInputAnswers("")
        self.assertEqual(tui.enterInt(), None)

        tui.addInputAnswers("a", "12")
        self.assertEqual(tui.enterInt(default=4), 12)

    def testSelectFromList(self):
        lst = [("a", "alpha"), ("b", "bravo"), ("c", "charlie")]
        tui.addInputAnswers("a")
        value = tui.selectFromList(lst, valueForString=str)
        self.assertEqual(value, "a")

        tui.addInputAnswers("z", "b")
        value = tui.selectFromList(lst, valueForString=str)
        self.assertEqual(value, "b")

    def testConfirm(self):
        tui.addInputAnswers("zog", "y")
        value = tui.confirm("bla")
        self.assertTrue(value)

        tui.addInputAnswers("zog", "n")
        value = tui.confirm("bla")
        self.assertFalse(value)



# -*- coding: UTF-8 -*-
"""
Database utilities.

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
from datetime import datetime, timedelta
import os

from sqlalchemy import and_
from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from yokadi.ycli import tui
from yokadi.core import db
from yokadi.core.db import Keyword, Project, Task, TaskKeyword, TaskLock
from yokadi.core.yokadiexception import YokadiException


def addTask(projectName, title, keywordDict=None, interactive=True):
    """Adds a task based on title and keywordDict.
    @param projectName: name of project as a string
    @param title: task title as a string
    @param keywordDict: dictionary of keywords (name : value)
    @param interactive: Ask user before creating project (this is the default)
    @type interactive: Bool
    @returns : Task instance on success, None if cancelled."""
    session = db.getSession()
    if keywordDict is None:
        keywordDict = {}

    # Create missing keywords
    if not createMissingKeywords(keywordDict.keys(), interactive=interactive):
        return None

    # Create missing project
    project = getOrCreateProject(projectName, interactive=interactive)
    if not project:
        return None

    # Create task
    task = Task(creationDate=datetime.now().replace(second=0, microsecond=0), project=project, title=title,
                description="", status="new")
    session.add(task)
    task.setKeywordDict(keywordDict)
    session.flush()

    return task


def getTaskFromId(tid):
    """Returns a task given its id, or raise a YokadiException if it does not
    exist.
    @param tid: Task id or uuid
    @return: Task instance or None if existingTask is False"""
    session = db.getSession()
    if isinstance(tid, str) and '-' in tid:
        filters = dict(uuid=tid)
    else:
        try:
            # We do not use line.isdigit() because it returns True if line is '¹'!
            taskId = int(tid)
        except ValueError:
            raise YokadiException("task id should be a number")
        filters = dict(id=taskId)

    try:
        task = session.query(Task).filter_by(**filters).one()
    except NoResultFound:
        raise YokadiException("Task %s does not exist. Use t_list to see all tasks" % taskId)
    return task


def getOrCreateKeyword(keywordName, interactive=True):
    """Get a keyword by its name. Create it if needed
    @param keywordName: keyword name as a string
    @param interactive: Ask user before creating keyword (this is the default)
    @type interactive: Bool
    @return: Keyword instance or None if user cancel creation"""
    session = db.getSession()
    try:
        return session.query(Keyword).filter_by(name=keywordName).one()
    except (NoResultFound, MultipleResultsFound):
        if interactive and not tui.confirm("Keyword '%s' does not exist, create it" % keywordName):
            return None
        keyword = Keyword(name=keywordName)
        session.add(keyword)
        print("Added keyword '%s'" % keywordName)
        return keyword


def getOrCreateProject(projectName, interactive=True, createIfNeeded=True):
    """Get a project by its name. Create it if needed
    @param projectName: project name as a string
    @param interactive: Ask user before creating project (this is the default)
    @type interactive: Bool
    @param createIfNeeded: create project if it does not exist (this is the default)
    @type createIfNeeded: Bool
    @return: Project instance or None if user cancel creation or createIfNeeded is False"""
    session = db.getSession()
    try:
        return session.query(Project).filter_by(name=projectName).one()
    except (NoResultFound, MultipleResultsFound):
        if not createIfNeeded:
            return None
        if interactive and not tui.confirm("Project '%s' does not exist, create it" % projectName):
            return None
        project = Project(name=projectName)
        session.add(project)
        print("Added project '%s'" % projectName)
        return project


def createMissingKeywords(lst, interactive=True):
    """Create all keywords from lst which does not exist
    @param lst: list of keyword
    @return: True, if ok, False if user canceled"""
    for keywordName in lst:
        if not getOrCreateKeyword(keywordName, interactive=interactive):
            return False
    return True


def getKeywordFromName(name):
    """Returns a keyword from its name, which may start with "@"
    raises a YokadiException if not found
    @param name: the keyword name
    @return: The keyword"""
    session = db.getSession()
    if not name:
        raise YokadiException("No keyword supplied")
    if name.startswith("@"):
        name = name[1:]
    lst = session.query(Keyword).filter_by(name=name).all()
    if len(lst) == 0:
        raise YokadiException("No keyword named '%s' found" % name)
    return lst[0]


def splitKeywordDict(dct):
    """Take a keyword dict and return a tuple of the form (userDict,
    reservedDict) """
    userDict = {}
    reservedDict = {}
    for key, value in dct.items():
        if key[0] == '_':
            reservedDict[key] = value
        else:
            userDict[key] = value
    return userDict, reservedDict


class TaskLockManager:
    """Handle a lock to prevent concurrent editing of the same task"""
    def __init__(self, task):
        """
        @param task: a Task instance
        @param session: sqlalchemy session"""
        self.task = task
        self.session = db.getSession()

    def _getLock(self):
        """Retrieve the task lock if it exists (else None)"""
        try:
            return db.getSession().query(TaskLock).filter(TaskLock.task == self.task).one()
        except NoResultFound:
            return None

    def acquire(self, pid=None, now=None):
        """Acquire a lock for that task and remove any previous stale lock"""
        if now is None:
            now = datetime.now()
        if pid is None:
            pid = os.getpid()

        lock = self._getLock()
        if lock:
            if lock.updateDate < now - 2 * timedelta(seconds=tui.MTIME_POLL_INTERVAL):
                # Stale lock, reusing it
                lock.pid = pid
                lock.updateDate = now
            else:
                raise YokadiException("Task %s is already locked by process %s" % (lock.task.id, lock.pid))
        else:
            # Create a lock
            self.session.add(TaskLock(task=self.task, pid=pid, updateDate=now))
        self.session.commit()

    def update(self, now=None):
        """Update lock timestamp to avoid it to expire"""
        if now is None:
            now = datetime.now()
        lock = self._getLock()
        lock.updateDate = now
        self.session.commit()

    def release(self):
        """Release the lock for that task"""
        # Only release our lock
        lock = self._getLock()
        if lock and lock.pid == os.getpid():
            self.session.delete(lock)
            self.session.commit()


class DbFilter(object):
    """
    Light wrapper around SQL Alchemy filters. Makes it possible to have the
    same interface as KeywordFilter
    """
    def __init__(self, condition):
        self.condition = condition

    def apply(self, lst):
        return lst.filter(self.condition)


class KeywordFilter(object):
    """Represent a filter on a keyword"""
    def __init__(self, name, negative=False, value=None, valueOperator=None):
        self.name = name  # Keyword name
        assert self.name, "Keyword name cannot be empty"
        self.negative = negative  # Negative filter
        self.value = value  # Keyword value
        self.valueOperator = valueOperator  # Operator to compare value

    def __repr__(self):
        return "<KeywordFilter name={} negative={} value={} valueOperator={}>".format(
            self.name, self.negative, self.value, self.valueOperator)

    def apply(self, query):
        """Apply keyword filters to query
        @return: a new query"""
        if self.negative:
            session = db.getSession()
            excludedTaskIds = session.query(Task.id).join(TaskKeyword).join(Keyword) \
                .filter(Keyword.name.like(self.name))
            return query.filter(~Task.id.in_(excludedTaskIds))
        else:
            keywordAlias = aliased(Keyword)
            taskKeywordAlias = aliased(TaskKeyword)
            query = query.outerjoin(taskKeywordAlias, Task.taskKeywords)
            query = query.outerjoin(keywordAlias, taskKeywordAlias.keyword)
            filter = keywordAlias.name.like(self.name)
            if self.valueOperator == "=":
                filter = and_(filter, taskKeywordAlias.value == self.value)
            elif self.valueOperator == "!=":
                filter = and_(filter, taskKeywordAlias.value != self.value)
            return query.filter(filter)

# vi: ts=4 sw=4 et



"""
Yokadi base class for test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os
import shutil
import tempfile
import unittest

from yokadi.tests.testutils import EnvironSaver


class YokadiTestCase(unittest.TestCase):
    """
    A TestCase which takes care of isolating the test from the user home dir
    and environment.
    """
    def setUp(self):
        self.__envSaver = EnvironSaver()
        self.testHomeDir = tempfile.mkdtemp(prefix="yokadi-basepaths-testcase")
        os.environ["HOME"] = self.testHomeDir
        os.environ["XDG_DATA_HOME"] = ""
        os.environ["XDG_CACHE_HOME"] = ""
        os.environ["YOKADI_DB"] = ""
        os.environ["YOKADI_HISTORY"] = ""
        self.__cwd = os.getcwd()

    def tearDown(self):
        shutil.rmtree(self.testHomeDir)
        self.__envSaver.restore()
        os.chdir(self.__cwd)



# -*- coding: UTF-8 -*-
"""
Implementation of completers for various Yokadi objects.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from yokadi.ycli.basicparseutils import parseParameters, simplifySpaces
from yokadi.core import db
from yokadi.core.db import Config, Keyword, Project, Task
from yokadi.core.ydateutils import WEEKDAYS
from yokadi.core.recurrencerule import FREQUENCIES


def computeCompleteParameterPosition(text, line, begidx, endidx):
    before = simplifySpaces(line[:begidx].strip())
    return before.count(" ") + 1


def getItemPropertiesStartingWith(item, field, text):
    """Return a list of item.field starting with text
    @param item: the object item, example : Task, Project, Keyword...
    @param field: the item's field lookup : Project.q.name, Task.q.title, Keyword.q.name. Don't forget the magic q
    @param text: The begining of the text as a str
    @return: list of matching strings"""
    session = db.getSession()
    return [x.name for x in session.query(item).filter(field.like(str(text) + "%"))]


class ProjectCompleter(object):
    def __init__(self, position):
        self.position = position

    def __call__(self, text, line, begidx, endidx):
        if computeCompleteParameterPosition(text, line, begidx, endidx) == self.position:
            return ["%s " % x for x in getItemPropertiesStartingWith(Project, Project.name, text)]
        else:
            return []


class KeywordCompleter(object):
    def __init__(self, position):
        self.position = position

    def __call__(self, text, line, begidx, endidx):
        if computeCompleteParameterPosition(text, line, begidx, endidx) == self.position:
            return getItemPropertiesStartingWith(Keyword, Keyword.name, text)
        else:
            return []


class MultiCompleter(object):
    """A completer which takes multiple completers and apply them in turn,
    according to their position"""
    def __init__(self, *completers):
        self.completers = completers

    def __call__(self, text, line, begidx, endidx):
        for completer in self.completers:
            lst = completer(text, line, begidx, endidx)
            if lst:
                return lst
        else:
            return []


def projectAndKeywordCompleter(cmd, text, line, begidx, endidx, shift=0):
    """@param shift: argument position shift. Used when command is omitted (t_edit usecase)"""
    position = computeCompleteParameterPosition(text, line, begidx, endidx)
    position -= len(parseParameters(line)[0])  # remove arguments from position count
    position += shift  # Apply argument shift
    if position == 1:  # Projects
        return ["%s" % x for x in getItemPropertiesStartingWith(Project, Project.name, text)]
    elif position >= 2 and line[-1] != " " and line.split()[-1][0] == "@":  # Keywords (we ensure that it starts with @
        return ["%s" % x for x in getItemPropertiesStartingWith(Keyword, Keyword.name, text)]


def confCompleter(cmd, text, line, begidx, endidx):
    return getItemPropertiesStartingWith(Config, Config.name, text)


def taskIdCompleter(cmd, text, line, begidx, endidx):
    # TODO: filter on parameter position
    # TODO: potential performance issue with lots of tasks, find a better way to do it
    session = db.getSession()
    tasks = [x for x in session.query(Task).filter(Task.status != 'done') if str(x.id).startswith(text)]
    print()
    for task in tasks:
        # Move that in a renderer class ?
        print("%s: %s / %s" % (task.id, task.project.name, task.title))
    return [str(x.id) for x in tasks]


def recurrenceCompleter(cmd, text, line, begidx, endidx):
    frequencies = [x.lower() for x in FREQUENCIES.values()] + ["none"]
    weekdays = [x.lower() for x in WEEKDAYS.keys()]

    position = computeCompleteParameterPosition(text, line, begidx, endidx)
    if position == 1:  # Task id
        return taskIdCompleter(cmd, text, line, begidx, endidx)
    elif position == 2:  # frequency
        return [x for x in frequencies if x.startswith(text.lower())]
    elif position == 3 and "weekly" in line.lower():
        return [x for x in weekdays if x.startswith(text.lower())]


def dueDateCompleter(cmd, text, line, begidx, endidx):
    position = computeCompleteParameterPosition(text, line, begidx, endidx)
    if position == 1:  # Task id
        return taskIdCompleter(cmd, text, line, begidx, endidx)
    elif position == 2 and not text.startswith("+"):  # week day
        return [str(x) for x in list(WEEKDAYS.keys()) if str(x).lower().startswith(text.lower())]

# vi: ts=4 sw=4 et



"""
Various file utility functions

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os


def createParentDirs(path, mode=0o777):
    parent = os.path.dirname(path)
    if os.path.exists(parent):
        return
    os.makedirs(parent, mode=mode)



# -*- coding: UTF-8 -*-
"""
Yokadi parser test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.core.yokadioptionparser import YokadiOptionParser
from yokadi.core.yokadiexception import YokadiException


class YokadiOptionParserTestCase(unittest.TestCase):

    def testQuote(self):
        parser = YokadiOptionParser()
        parser.add_argument("cmd", nargs='*')
        src = "There's a quote here"
        args = parser.parse_args(src)
        # Recreate the line
        line = " ".join(args.cmd)
        self.assertEqual(line, src)

    def testDash(self):
        parser = YokadiOptionParser()
        parser.add_argument("cmd", nargs="*")
        srcs = ["foo-bar", "foo - bar"]
        for src in srcs:
            args = parser.parse_args(src)
            # Recreate the line
            line = " ".join(args.cmd)
            self.assertEqual(line, src)

    def testUknownOption(self):
        def parseUnknownOption():
            parser.parse_args("blabla -b")
        parser = YokadiOptionParser()
        self.assertRaises(YokadiException, parseUnknownOption)

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Keyword related commands.

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
from sqlalchemy.exc import IntegrityError

from yokadi.core import dbutils
from yokadi.ycli import tui

from yokadi.core import db
from yokadi.core.db import Keyword
from yokadi.core.yokadiexception import BadUsageException
from yokadi.ycli.completers import KeywordCompleter


def _listKeywords(session):
    for keyword in sorted(session.query(Keyword).all(), key=lambda x: x.name.lower()):
        taskIds = sorted([x.id for x in keyword.tasks if x])
        yield keyword.name, taskIds


class KeywordCmd(object):
    def do_k_list(self, line):
        """List all keywords."""
        for name, taskIds in _listKeywords(db.getSession()):
            if taskIds:
                tasks = ", ".join([str(x) for x in taskIds])
            else:
                tasks = "none"
            print("{} (tasks: {})".format(name, tasks))

    def do_k_add(self, line):
        """Add a keyword
        k_add @<keyword1> [@<keyword2>...]"""
        session = db.getSession()
        if not line:
            raise BadUsageException("You must provide at least one keyword name")
        for keyword in line.split():
            try:
                session.add(Keyword(name=keyword))
                session.commit()
                print("Keyword %s has been created" % keyword)
            except IntegrityError:
                session.rollback()
                print("Keyword %s already exist" % keyword)

    def do_k_remove(self, line):
        """Remove a keyword
        k_remove @<keyword>"""
        session = db.getSession()
        keyword = dbutils.getKeywordFromName(line)

        if keyword.tasks:
            taskList = ", ".join(str(task.id) for task in keyword.tasks)
            print("The keyword {} is used by the following tasks: {}".format(keyword.name, taskList))
            if not tui.confirm("Do you really want to remove this keyword"):
                return
        session.delete(keyword)
        session.commit()
        print("Keyword {} has been removed".format(keyword.name))

    complete_k_remove = KeywordCompleter(1)

    def do_k_edit(self, line):
        """Edit a keyword
        k_edit @<keyword>"""
        session = db.getSession()
        keyword = dbutils.getKeywordFromName(line)
        oldName = keyword.name
        newName = tui.editLine(oldName)
        if newName == "":
            print("Cancelled")
            return

        lst = session.query(Keyword).filter_by(name=newName).all()
        if len(lst) == 0:
            # Simple case: newName does not exist, just rename the existing keyword
            keyword.name = newName
            session.commit()
            print("Keyword %s has been renamed to %s" % (oldName, newName))
            return

        # We already have a keyword with this name, we need to merge
        print("Keyword %s already exists" % newName)
        if not tui.confirm("Do you want to merge %s and %s" % (oldName, newName)):
            return

        # Check we can merge
        conflictingTasks = []
        for task in keyword.tasks:
            kwDict = task.getKeywordDict()
            if oldName in kwDict and newName in kwDict and kwDict[oldName] != kwDict[newName]:
                conflictingTasks.append(task)

        if len(conflictingTasks) > 0:
            # We cannot merge
            tui.error("Cannot merge keywords %s and %s because they are both"
                      " used with different values in these tasks:" % (oldName, newName))
            for task in conflictingTasks:
                print("- %d, %s" % (task.id, task.title))
            print("Edit these tasks and try again")
            return

        # Merge
        for task in keyword.tasks:
            kwDict = task.getKeywordDict()
            if newName not in kwDict:
                kwDict[newName] = kwDict[oldName]
            del kwDict[oldName]
            task.setKeywordDict(kwDict)
        session.delete(keyword)
        session.commit()
        print("Keyword %s has been merged with %s" % (oldName, newName))

    complete_k_edit = KeywordCompleter(1)



# -*- coding: UTF-8 -*-
"""
A simple exception class for Yokadi

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""


class YokadiException(Exception):
    """Yokadi Exceptions"""
    pass


class BadUsageException(YokadiException):
    """Exception when user does not pass correct arguments to a command"""
    pass
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Project test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

import testutils

from yokadi.core import db, dbutils
from yokadi.core.db import Project, Task
from yokadi.core.yokadiexception import YokadiException
from yokadi.ycli.main import YokadiCmd
from yokadi.ycli import tui


class ProjectTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        tui.clearInputAnswers()
        self.cmd = YokadiCmd()

    def testAdd(self):
        tui.addInputAnswers("y")
        self.cmd.do_p_add("p1")
        self.cmd.do_p_add("p2")

        projects = self.session.query(Project).all()
        result = [x.name for x in projects]
        expected = ["p1", "p2"]
        self.assertEqual(result, expected)

    def testEdit(self):
        # Create project p1 and rename it to p2
        self.cmd.do_p_add("p1")
        project = self.session.query(Project).filter_by(id=1).one()
        self.assertEqual(project.name, "p1")

        tui.addInputAnswers("p2")
        self.cmd.do_p_edit("p1")
        self.assertEqual(project.name, "p2")

        # Create project p3 and try to rename it to p2
        self.cmd.do_p_add("p3")
        project = self.session.query(Project).filter_by(name="p3").one()
        self.assertEqual(project.name, "p3")

        tui.addInputAnswers("p2")
        self.assertRaises(YokadiException, self.cmd.do_p_edit, "p3")
        self.assertEqual(project.name, "p3")

    def testRemove(self):
        # Create project p1 with one associated task
        tui.addInputAnswers("y")
        self.cmd.do_p_add("p1")
        self.session.query(Project).one()
        task = dbutils.addTask("p1", "t1", interactive=False)
        taskId = task.id

        # Remove project, its task should be removed
        tui.addInputAnswers("y")
        self.cmd.do_p_remove("p1")

        self.assertEqual(list(self.session.query(Task).filter_by(id=taskId)), [])

    def testStatus(self):
        # Create project p1 and test set active and set inactive method
        self.cmd.do_p_add("p1")
        project = self.session.query(Project).filter_by(id=1).one()
        self.assertEqual(project.name, "p1")

        self.assertEqual(project.active, True)
        self.cmd.do_p_set_inactive("p1")
        self.assertEqual(project.active, False)
        self.cmd.do_p_set_active("p1")
        self.assertEqual(project.active, True)

    def testMerge(self):
        COUNT = 4
        for x in range(COUNT):
            dbutils.addTask('p1', 'p1-t{}'.format(x), interactive=False)
            dbutils.addTask('p2', 'p2-t{}'.format(x), interactive=False)

        # Merge p1 into p2
        tui.addInputAnswers("y")
        self.cmd.do_p_merge("p1 p2")

        # p2 should have both its tasks and all p1 tasks now
        project = self.session.query(Project).filter_by(name="p2").one()
        tasks = set([x.title for x in project.tasks])

        expected = set()
        for x in range(COUNT):
            expected.add('p1-t{}'.format(x))
            expected.add('p2-t{}'.format(x))
        self.assertEqual(tasks, expected)

        # p1 should be gone
        testutils.assertQueryEmpty(self, self.session.query(Project).filter_by(name="p1"))

    def testMergeItselfFails(self):
        project = Project(name="p1")
        self.assertRaises(YokadiException, project.merge, self.session, project)


# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Helper functions to build CLI applications

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import os
import readline
import subprocess
import sys
import tempfile
import time
import unicodedata
import re
import locale
import shutil
from collections import namedtuple
from getpass import getpass

from yokadi.ycli import colors
from yokadi.core.yokadiexception import YokadiException

# Number of seconds between checks for end of process
PROC_POLL_INTERVAL = 0.5
# Number of seconds between checks for file modification
MTIME_POLL_INTERVAL = 10

# Filter out bad characters for filenames
NON_SIMPLE_ASCII = re.compile("[^a-zA-Z0-9]+")
MULTIPLE_DASH = re.compile("-+")

_answers = []

stdout = sys.stdout
stderr = sys.stderr


_isInteractive = sys.stdin.isatty()


def isInteractive():
    if _answers:
        # We are in a test, interaction is being simulated
        return True
    return _isInteractive


def _checkIsInteractive():
    if not isInteractive():
        raise YokadiException("This command cannot be used in non-interactive mode")


def editText(text, onChanged=None, lockManager=None, prefix="yokadi-", suffix=".md"):
    """Edit text with external editor
    @param onChanged: function parameter that is call whenever edited data change. Data is given as a string
    @param lockManager: function parameter that is called to 'acquire', 'update' or 'release' an editing lock
    @param prefix: temporary file prefix.
    @param suffix: temporary file suffix.
    @return: newText"""
    _checkIsInteractive()
    encoding = locale.getpreferredencoding()

    def readFile(name):
        with open(name, encoding=encoding) as data:
            return str(data.read())

    def waitProcess(proc):
        start = time.time()
        while (time.time() - start) < MTIME_POLL_INTERVAL:
            proc.poll()
            if proc.returncode is not None:
                return
            time.sleep(PROC_POLL_INTERVAL)

    prefix = NON_SIMPLE_ASCII.sub("-", prefix)
    prefix = MULTIPLE_DASH.sub("-", prefix)
    prefix = unicodedata.normalize('NFKD', prefix)

    (fd, name) = tempfile.mkstemp(suffix=suffix, prefix=prefix)
    if text is None:
        text = ""
    try:
        if lockManager:
            lockManager.acquire()
        fl = open(name, "w", encoding=encoding)
        fl.write(text)
        fl.close()
        editor = os.environ.get("EDITOR", "vi")
        proc = subprocess.Popen([editor, name])
        mtime = os.stat(name).st_mtime
        while proc.returncode is None:
            waitProcess(proc)
            if proc.returncode is None and lockManager is not None:
                lockManager.update()
            if proc.returncode is None and onChanged is not None:
                newMtime = os.stat(name).st_mtime
                if newMtime > mtime:
                    mtime = newMtime
                    onChanged(readFile(name))
        if proc.returncode != 0:
            raise Exception("The command {} failed. It exited with code {}.".format(proc.args, proc.returncode))
        return readFile(name)
    finally:
        os.close(fd)
        os.unlink(name)
        if lockManager:
            lockManager.release()


def reinjectInRawInput(line):
    """Next call to input() will have line set as default text
    @param line: The default text
    """
    assert isinstance(line, str)

    # Set readline.pre_input_hook to feed it with our line
    # (Code copied from yagtd)
    def pre_input_hook():
        readline.insert_text(line)
        readline.redisplay()

        # Unset the hook again
        readline.set_pre_input_hook(None)

    if sys.platform != 'win32':
        readline.set_pre_input_hook(pre_input_hook)


def editLine(line, prompt="edit> ", echo=True):
    """Edit a line using readline
    @param prompt: change prompt
    @param echo: whether to echo user text or not"""
    _checkIsInteractive()
    if line:
        reinjectInRawInput(line)

    if len(_answers) > 0:
        line = _answers.pop(0)
    else:
        try:
            if echo:
                line = input(prompt)
            else:
                line = getpass(prompt)
        except EOFError:
            line = ""

    # Remove edited line from history:
    #   oddly, get_history_item is 1-based,
    #   but remove_history_item is 0-based
    if sys.platform != 'win32':
        length = readline.get_current_history_length()
        if length > 0:
            readline.remove_history_item(length - 1)

    return line


def selectFromList(lst, default=None, prompt="Select", valueForString=int):
    """
    Takes a list of tuples (value, caption), returns the value of the selected
    entry.
    @param default indicates the default value and may be None
    @param prompt customize the prompt
    @param valueForString a function to turn a string into a valid value
    """
    if default is not None:
        default = str(default)
    possibleValues = {x[0] for x in lst}
    for value, caption in lst:
        print("{}: {}".format(value, caption))

    while True:
        line = editLine(default, prompt=prompt + ": ")
        try:
            value = valueForString(line)
        except Exception:
            error("Wrong value")
            continue

        if value in possibleValues:
            return value
        else:
            error("Wrong value")


def enterInt(default=None, prompt="Enter a number"):
    if default is None:
        line = ""
    else:
        line = str(default)
    while True:
        answer = editLine(line, prompt=prompt + ": ")
        if answer == "":
            return None
        try:
            value = int(answer)
            return value
        except ValueError:
            error("Wrong value")


def confirm(prompt):
    if not isInteractive():
        return True
    while True:
        answer = editLine("", prompt=prompt + " (y/n)? ")
        answer = answer.lower()

        if answer == "y":
            return True
        elif answer == "n":
            return False
        else:
            error("Wrong value")


def renderFields(fields):
    """Print on screen tabular array represented by fields
    @param fields: list of tuple (caption, value)
    """
    maxWidth = max([len(x) for x, y in fields])
    format = colors.BOLD + "%" + str(maxWidth) + "s" + colors.RESET + ": %s"
    for caption, value in fields:
        print(format % (caption, value), file=stdout)


def warnDeprecated(old, new):
    """Warn user that a command is now deprecated
    and incitate him to use the new one
    @param old: the old one (str)
    @param new: the new one (str)"""
    warning("Command '%s' is deprecated, use '%s' instead" % (old, new))
    info("Command %s has been executed" % new)


def error(message):
    print(colors.BOLD + colors.RED + "Error: %s" % message + colors.RESET, file=stderr)


def warning(message):
    print(colors.RED + "Warning: " + colors.RESET + message, file=stderr)


def info(message):
    print(colors.CYAN + "Info: " + colors.RESET + message, file=stderr)


def addInputAnswers(*answers):
    """Add answers to tui internal answer buffer. Next call to editLine() will
    pop the first answer from the buffer instead of prompting the user.
    This is useful for unit-testing."""
    _answers.extend(answers)


def clearInputAnswers():
    """Remove all added answers. Useful to avoid making a test depend on a "y"
    added by another test.
    """
    global _answers
    _answers = []


def getTermWidth():
    """Gets the terminal width"""
    size = shutil.get_terminal_size()
    return size.columns


ColorBlock = namedtuple("ColorBlock", ("pos", "color"))


class TextColorizer:
    def __init__(self):
        self._dct = {}

    def setColorAt(self, pos, color):
        self._dct[pos] = color

    def setResetAt(self, pos):
        self._dct[pos] = colors.RESET

    def crop(self, width):
        self._dct = {pos: color for pos, color in self._dct.items() if pos < width}

    def render(self, text):
        """
        Apply color blocks to text
        """
        start = 0
        out = []
        blockList = sorted(self._dct.items())
        for pos, color in blockList:
            out.append(text[start:pos] + color)
            start = pos
        # Add remaining text, if any
        out.append(text[start:])
        return "".join(out)
# vi: ts=4 sw=4 et
